#!/usr/bin/env python3

# pylint: disable=protected-access

import os

from unittest import TestCase

import radical.utils as ru


# ------------------------------------------------------------------------------
#
class ConfigTestCase(TestCase):

    # --------------------------------------------------------------------------
    #
    def test_config(self):

        base = os.path.abspath(os.path.dirname(__file__))
        cfg_files = '%s/data/resource_*.json' % base

        with self.assertRaises(ValueError):
            ru.Config(path='foo', cfg='bar')

        cfg1 = ru.Config(name=cfg_files)
        self.assertEqual(cfg1.yale._query('grace.agent_launch_method'), 'bar')
        self.assertEqual(cfg1['yale']['grace']['agent_launch_method'],  'bar')
        self.assertIsNone(cfg1._query('yale.grace.no_launch_method'))
        with self.assertRaises(KeyError):
            # key is not set
            _ = cfg1['yale']['grace']['no_launch_method']            # noqa F841

        os.environ['FOO'] = 'GSISSH'
        # `agent_launch_method` is defined by env variable `$FOO`

        cfg2 = ru.Config(name=cfg_files)
        self.assertEqual(cfg2.yale.grace.agent_launch_method, 'GSISSH')

        # env variables are not expanded
        cfg3 = ru.Config(name=cfg_files, expand=False)
        self.assertEqual(cfg3.yale.grace.agent_launch_method, '${FOO:bar}')

        env  = {'FOO': 'baz'}
        cfg4 = ru.Config(name=cfg_files, env=env)
        self.assertEqual(cfg4['yale']['grace']['agent_launch_method'], 'baz')

        # test `cls._self_default` flag
        cfg5 = ru.Config(from_dict={'foo_0': {'foo_1': {'foo2': 'bar'}}})
        self.assertEqual(cfg5.foo_0.foo_1.foo2, 'bar')
        self.assertIsInstance(cfg5.foo_0.foo_1, ru.Config)

    # --------------------------------------------------------------------------
    #
    def test_default_config(self):

        dc1 = ru.DefaultConfig()
        self.assertIn(type(dc1), ru.Singleton._instances)
        self.assertEqual(dc1, ru.Singleton._instances[type(dc1)])

        c = ru.Config(module='radical.utils', category='utils', name='default')
        self.assertEqual(c.as_dict(), dc1.as_dict())

        # do not change other attributes for other tests consistency
        dc1.report_dir = '/tmp'

        dc2 = ru.DefaultConfig()  # check that there is ony one instance
        self.assertEqual(id(dc1), id(dc2))
        self.assertEqual(dc1, dc2)

# ------------------------------------------------------------------------------




# pylint: disable=protected-access

__author__    = 'RADICAL-Cybertools Team'
__copyright__ = 'Copyright 2021-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

from unittest import TestCase

from radical.utils.typeddict import TDError, TDKeyError
from radical.utils.typeddict import TDTypeError, TDValueError
from radical.utils           import TypedDict


# ------------------------------------------------------------------------------
#
class TypedDictTestCase(TestCase):

    # --------------------------------------------------------------------------
    #
    @classmethod
    def setUpClass(cls) -> None:

        class TDEmpty(TypedDict):
            pass

        class TDSimple(TypedDict):

            _schema = {
                'attr_str' : str,
                'attr_int' : int,
                'attr_dict': {str: None}
            }

            _defaults = {
                'attr_str' : 'string_default',
                'attr_int' : '1',
                'attr_dict': {'dict_key_00': 100}
            }

        class TDDoubleLevel(TypedDict):

            _schema = {
                'id'       : int,
                'simple_td': TDSimple
            }

        # ----------------------------------------------------------------------

        cls.test_td_cases = [
            (TDEmpty,       {'attr_any_00': 'attr_00',
                             'attr_any_01': 'attr_01'}),
            (TDSimple,      {'attr_str' : 'non_default',
                             'attr_int' : 2,
                             'attr_dict': {'dict_key_00': 111}}),
            (TDDoubleLevel, {'id'       : 3,
                             'simple_td': {'attr_str' : 'simple_td_str',
                                           'attr_int' : 10}})
        ]

    # --------------------------------------------------------------------------
    #
    def test_init(self):

        def _assertion_in_loop(td_obj_value, input_value):
            if isinstance(input_value, (dict, TypedDict)):
                for _k, _v in input_value.items():
                    self.assertIn(_k, td_obj_value)
                    _assertion_in_loop(td_obj_value[_k], _v)
            elif isinstance(input_value, (list, tuple)):
                self.assertTrue(isinstance(td_obj_value, (list, tuple)))
                for _idx in range(len(input_value)):
                    _assertion_in_loop(td_obj_value[_idx],
                                       input_value[_idx])
            else:
                self.assertEqual(td_obj_value, input_value)

        # ----------------------------------------------------------------------

        for cls, input_data in self.test_td_cases:

            obj = cls()

            if obj._schema and obj._defaults:

                self.assertEqual(len(obj), len(obj._defaults))

                for k, v in cls._defaults.items():
                    self.assertEqual(obj[k], v)

                obj.verify()
                for k in obj.keys():
                    if isinstance(cls._schema[k], type):
                        t = cls._schema[k]
                    else:
                        t = type(cls._schema[k])
                    self.assertIsInstance(obj[k], t)

                obj.random_attr = 'random'
                with self.assertRaises(KeyError):
                    # due to "random_attr" is not in schema
                    obj.verify()

            elif not obj._schema:

                self.assertEqual(len(obj), 0)

                # add one element
                obj.random_attr = 'random'
                self.assertEqual(len(obj), 1)
                self.assertIn('random_attr', obj.keys())
                # verification will be skipped since there is no "_schema"
                obj.verify()

            if input_data:

                obj = cls(from_dict=input_data)
                _assertion_in_loop(td_obj_value=obj, input_value=input_data)

    # --------------------------------------------------------------------------
    #
    def test_hash(self):

        class TDHashed(TypedDict):

            def __hash__(self):
                return hash(tuple(sorted(self.items())))

        # ----------------------------------------------------------------------

        obj = TypedDict()

        with self.assertRaises(TypeError) as e:
            hash(obj)
        self.assertIn('unhashable type', str(e.exception))

        input_data = {'a': 'a', 'b': 'b'}
        self.assertEqual(hash(TDHashed(input_data)), hash(TDHashed(input_data)))

    # --------------------------------------------------------------------------
    #
    def test_self_default(self):

        # test `cls._self_default` flag (for method `update`)

        class TDNonSelf(TypedDict):

            _self_default = False

        class TDSelf(TypedDict):

            _self_default = True

        # ----------------------------------------------------------------------

        obj1 = TDNonSelf(from_dict={'level0': {'level1': {'level2': 'value'}}})
        self.assertIsInstance(obj1, TDNonSelf)
        self.assertNotIsInstance(obj1.level0, TDNonSelf)
        self.assertIsInstance(obj1.level0.level1, TypedDict)

        obj2 = TDSelf(from_dict={'level0': {'level1': {'level2': 'value'}}})
        self.assertIsInstance(obj2, TDSelf)
        self.assertIsInstance(obj2.level0, TDSelf)
        self.assertTrue(issubclass(type(obj2.level0.level1), TypedDict))

        # all instances are treated as `dict`
        self.assertIsInstance(obj1, dict)
        self.assertIsInstance(obj2, dict)

    # --------------------------------------------------------------------------
    #
    def test_verify(self):

        # attribute "_cast" is True by default

        class TDBase(TypedDict):

            _schema = {
                'attr_int' : int,
                'attr_dict': {str: int}
            }

        class TDMain(TypedDict):

            _schema = {
                'attr_main_int': int,
                'attr_main_td' : TDBase
            }

        class TDBaseCopy(TypedDict):

            _schema = {
                'attr_int' : int,
                'attr_dict': {str: int}
            }

        # ----------------------------------------------------------------------

        input_data = {'attr_main_int': 2,
                      'attr_main_td' : {'attr_int' : 3,
                                        'attr_dict': {'attr_any': 2}}}
        obj = TDMain(input_data)
        obj.verify()
        self.assertEqual(type(obj.attr_main_td).__name__, 'TDBase')
        self.assertEqual(obj.as_dict(), input_data)

        with self.assertRaises(TypeError):
            # will not be able to convert attribute "attr_main_int"
            TDMain({'attr_main_int': 'non_int',
                    'attr_main_td' : {'attr_int' : 3,
                                      'attr_dict': {'attr_any': 2}}}).verify()

        with self.assertRaises(KeyError):
            # will not be able to convert attribute "attr_main_td"
            TDMain({'attr_main_int': 2,
                    'attr_main_td' : {'not_in_schema': 3}}).verify()

        with self.assertRaises(KeyError):
            # will not be able to convert input data for `TDBase`
            TDMain({'attr_main_int': 2,
                    'attr_main_td' : TDBase({'not_in_schema': 3})}).verify()

        with self.assertRaises(TypeError):
            # will not be able to convert attribute "attr_main_td.attr_dict"
            TDMain({'attr_main_int': 2,
                    'attr_main_td' : {'attr_int' : 3,
                                      'attr_dict': {'attr_any': 'a'}}}).verify()

        obj = TDMain({'attr_main_int': 2,
                      'attr_main_td' : TDBaseCopy()})
        # no TDError, since value of attribute "attr_main_td" will be converted
        # classes `TDBase` and `TDBaseCopy` have the same schemas
        obj.verify()

    # --------------------------------------------------------------------------
    #
    def test_verify_setter(self):

        class TDBaseNoSchema(TypedDict):

            _check = True

        class TDBase(TypedDict):

            _check = True

            _schema = {
                'attr_str'  : str,
                'attr_float': float
            }

        # ----------------------------------------------------------------------

        # without set "_schema"
        obj_wo_schema = TDBaseNoSchema()

        obj_wo_schema.attr_str = 'str_value'
        self.assertIsInstance(obj_wo_schema.attr_str, str)
        obj_wo_schema.attr_dict = {}
        self.assertIsInstance(obj_wo_schema.attr_dict, dict)

        # with provided "_schema"
        obj = TDBase()

        with self.assertRaises(KeyError):
            # provided attribute is not in schema
            obj.attr_dict = {}

        # attribute "_cast" is True by default
        obj.attr_str = 1
        self.assertIsInstance(obj.attr_str, str)
        self.assertEqual(obj.attr_str, '1')
        obj.attr_float = 1
        self.assertIsInstance(obj.attr_float, float)
        self.assertEqual(obj.attr_float, 1.)
        with self.assertRaises(TypeError):
            # couldn't convert provided value into a corresponding type (float)
            obj.attr_float = 'str_value'

        # set "_cast" to False
        TDBase._cast = False
        obj.attr_str = 'new_str_value'
        self.assertIsInstance(obj.attr_str, str)
        with self.assertRaises(TypeError):
            # no attempts to convert between types
            obj.attr_str = 1

    # --------------------------------------------------------------------------
    #
    def test_base_methods(self):

        input_dict = {'attr_any': 'attr_any_str'}
        obj = TypedDict(from_dict=input_dict)

        self.assertEqual(obj.keys(), input_dict.keys())

        for k in obj:
            self.assertIn(k, input_dict)

        for k in obj.keys():
            self.assertIn(k, input_dict)

        for v in obj.values():
            self.assertIn(v, input_dict.values())

        for k, v in obj.items():
            self.assertIn(k, input_dict)
            self.assertEqual(v, input_dict[k])

        k = list(input_dict)[0]

        self.assertEqual(obj.get(k), input_dict[k])
        self.assertEqual(obj.get('non_set_attr', 'default_v'), 'default_v')

        self.assertEqual(obj.setdefault(k), input_dict[k])
        pre_setdefault_len = len(obj)
        self.assertEqual(obj.setdefault('new_attr', 'new_value'), 'new_value')
        self.assertIn('new_attr', obj)
        self.assertEqual(len(obj), pre_setdefault_len + 1)

        obj.clear()
        self.assertEqual(len(obj), 0)

        class TDSchemed(TypedDict):

            _schema = {
                'attr_str'  : str,
                'attr_float': float
            }

        tds = TDSchemed()
        self.assertIsNone(tds.attr_str)
        with self.assertRaises(KeyError):
            # not defined earlier and not from schema
            _ = tds.unknown_key

        tds.unknown_key = 'unknown_key_value'
        self.assertEqual(tds.unknown_key, 'unknown_key_value')

        del tds.unknown_key
        self.assertNotIn('unknown_key', tds)

        # `__str__` method checked
        self.assertEqual('%s' % tds, '{}')
        # `__repr__` method checked
        self.assertIn('TDSchemed object, schema keys', '%r' % tds)

    # --------------------------------------------------------------------------
    #
    def test_pop(self):

        for cls, _ in self.test_td_cases:

            obj = cls()
            if obj._schema and obj._defaults:

                initial_obj_keys = list(obj.keys())
                for k in initial_obj_keys:
                    v = obj.pop(k)
                    self.assertIsNotNone(v)
                    self.assertNotIn(k, obj)

            v = 'random_value'
            self.assertEqual(v, obj.pop('unknown_key', v))

            with self.assertRaises(KeyError) as e:
                obj.pop('unknown_key')
            self.assertIn('not found', str(e.exception))

    # --------------------------------------------------------------------------
    #
    def test_popitem(self):

        for cls, _ in self.test_td_cases:

            obj = cls()
            if obj._schema and obj._defaults:

                num_elements = len(obj)
                for _ in range(num_elements):
                    k, v = obj.popitem()
                    self.assertIsNotNone(v)
                    self.assertNotIn(k, obj)

            with self.assertRaises(TDError) as e:
                obj.popitem()
            self.assertIn('no data', str(e.exception))

    # --------------------------------------------------------------------------
    #
    def test_query(self):

        class TDLowLevel(TypedDict):

            _schema = {
                'low_dict': {str: None},
                'low_str' : str
            }

            _defaults = {
                'low_dict': {'low_dict_int': 1},
                'low_str' : 'low_any_string'
            }

        class TDHighLevel(TypedDict):

            _schema = {
                'high_dict': dict,
                'high_td'  : TDLowLevel,
                'high_str' : str
            }

        # ----------------------------------------------------------------------

        td_obj = TDHighLevel({
            'high_dict': {
                'high_dict_l1': {
                    'high_dict_str': 'l2_output'
                }
            },
            'high_td' : TDLowLevel(),
            'high_str': 'high_any_string'
        })

        self.assertIsInstance(
            td_obj._query('high_dict.high_dict_l1.high_dict_str'), str)
        self.assertIsInstance(
            td_obj._query(['high_dict', 'high_dict_l1', 'high_dict_str']), str)
        self.assertIsInstance(
            td_obj._query('high_str'), str)
        self.assertIsInstance(
            td_obj._query('high_td.low_str'), str)
        self.assertIsInstance(
            td_obj._query('high_td.low_dict'), dict)
        self.assertIsInstance(
            td_obj._query('high_td.low_dict.low_dict_int'), int)

        self.assertIsNone(td_obj._query('high_dict.high_dict_l1.unknown_key'))
        self.assertIsNone(td_obj._query('high_td.low_dict.unknown_key'))

        self.assertIsInstance(td_obj._query('high_td.low_dict.unknown_key',
                                            default=1), int)
        self.assertIsInstance(td_obj._query('high_td.unknown_sub_key.levelK',
                                            default=1,
                                            last_key=False), int)

        with self.assertRaises(KeyError):
            td_obj._query('')

        with self.assertRaises(TDKeyError):
            td_obj._query(None)

        with self.assertRaises(ValueError):
            td_obj._query('high_dict.high_dict_l1.high_dict_str.within_non_dict')

        with self.assertRaises(TDValueError):
            td_obj._query('high_td.low_dict.low_dict_int.within_non_dict')

        with self.assertRaises(KeyError):
            td_obj._query('high_td.unknown_sub_key.levelK')

    # --------------------------------------------------------------------------
    #
    def test_metaclass(self):

        class TD1Base(TypedDict):

            _self_default = True

            _schema = {
                'base_int': int,
                'base_str': str
            }

        class TD2Base(TD1Base):

            _cast = False

            _schema = {
                'base_int': float
            }

            _defaults = {
                'base_int': .5
            }

        class TD3Base(TD2Base):
            pass

        # ----------------------------------------------------------------------

        # check that "base" attributes are provided
        for base_attr in ['_schema',
                          '_defaults',
                          '_self_default',
                          '_check',
                          '_cast']:
            self.assertTrue(hasattr(TD1Base, base_attr))
            self.assertTrue(hasattr(TD2Base, base_attr))
            self.assertTrue(hasattr(TD3Base, base_attr))

        # inherited "_defaults" from TD2Base
        self.assertEqual(getattr(TD3Base, '_defaults')['base_int'], .5)

        # inherited "_schema" from TD2Base and TD1Base
        self.assertIs(getattr(TD3Base, '_schema')['base_int'], float)
        self.assertIs(getattr(TD2Base, '_schema')['base_int'], float)
        self.assertIs(getattr(TD1Base, '_schema')['base_int'], int)

        # inherited "_self_default" from TD1Base (default value is False)
        self.assertTrue(getattr(TD3Base, '_self_default'))

        # inherited "_cast" from TD2Base
        self.assertFalse(getattr(TD3Base, '_cast'))
        self.assertFalse(getattr(TD2Base, '_cast'))
        # inherited from TypedDict (TD1Base class didn't change this attribute)
        self.assertTrue(getattr(TD1Base, '_cast'))

        # inherited from TD1Base ("_schema")
        td3 = TD3Base({'base_int': 10, 'base_str': 20})
        # exception due to `TD3Base._cast = False` (inherited from TD2Base)
        with self.assertRaises(TypeError):
            td3.verify()
        with self.assertRaises(TDTypeError) as e:
            td3.verify()
        self.assertIn('attribute "base_int" - expected type', str(e.exception))
        # NOTE: control flags should be set through the class only
        TD3Base._cast = True
        td3.verify()
        self.assertIsInstance(td3.base_int, float)
        self.assertIsInstance(td3.base_str, str)
        self.assertEqual(td3.base_int, 10.)
        self.assertEqual(td3.base_str, '20')

    # --------------------------------------------------------------------------
    #
    def test_tderrors(self):

        def raise_exception(level=1):
            raise TDError('level %s' % level, level=level)

        def raise_exception_wrapper(level):
            raise_exception(level=level)

        class TDErrorClass:

            @classmethod
            def raise_cls_tderror(cls):
                raise TDError

            def raise_self_tderror(self):
                raise TDError

        # ----------------------------------------------------------------------

        with self.assertRaises(TDError) as e:
            raise_exception()
        self.assertIn('.raise_exception - level 1', str(e.exception))
        # no class or instance provided
        self.assertIn('<>', str(e.exception))

        with self.assertRaises(TDError) as e:
            raise_exception_wrapper(level=1)
        self.assertIn('.raise_exception - level 1', str(e.exception))

        with self.assertRaises(TDError) as e:
            raise_exception_wrapper(level=2)
        self.assertIn('.raise_exception_wrapper - level 2', str(e.exception))

        # exception raised within class/instance methods

        with self.assertRaises(TDError) as e:
            # `'cls' in f.f_locals` -> True
            TDErrorClass.raise_cls_tderror()
        self.assertIn('TDErrorClass', str(e.exception))

        with self.assertRaises(TDError) as e:
            # `'self' in f.f_locals` -> True
            TDErrorClass().raise_self_tderror()
        self.assertIn('TDErrorClass', str(e.exception))

        # catch exceptions by their parent classes

        with self.assertRaises(Exception):
            raise TDError

        with self.assertRaises(KeyError):
            raise TDKeyError

        with self.assertRaises(TypeError):
            raise TDTypeError

        with self.assertRaises(ValueError):
            raise TDValueError

# ------------------------------------------------------------------------------



#!/usr/bin/env python

import os
import time
import socket

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_poll():

    sp     = socket.socketpair(socket.AF_UNIX, socket.SOCK_STREAM, 0)
    child  = os.fork()

    if not child:

        sp[0].close()
        sp[1].send(b'foo')
        time.sleep(0.1)
        sp[1].send(b'bar')
        time.sleep(0.2)
        sp[1].close()
        return

    else:

        sp[1].close()

        poller = ru.Poller()
        poller.register(sp[0], ru.POLLERR | ru.POLLIN | ru.POLLHUP)

        abort = False
        msgs  = list()
        while not abort:

            events = poller.poll(1.0)

            for _,event in events:

                if event & ru.POLLIN:
                    msg = sp[0].recv(200)
                    assert(msg in [b'foo', b'bar', b''])
                    msgs.append(msg)

                if  event & ru.POLLERR or event & ru.POLLHUP:
                    abort = True
                    assert(msgs[0] == b'foo')
                    assert(msgs[1] == b'bar')
                    assert(msgs[2] == b'')
                    break

        os.waitpid(child, 0)


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_poll()


# ------------------------------------------------------------------------------





#!/usr/bin/env python

__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import os
import copy
import pytest
import tempfile

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_import_file():

    syms = ru.import_file('%s/data/import_file.py' % os.path.dirname(__file__))

    assert(syms['functions']['foo'](1))
    assert(syms['functions']['foo'](4, 4))

    f = syms['classes']['Foo']()
    assert(f.foo(1))
    assert(f.foo(4, 4))


# ------------------------------------------------------------------------------
#
def test_round_to_base():

    assert(ru.round_to_base(1.5, 2) == 2)
    assert(ru.round_to_base(3.5, 2) == 4)
    assert(ru.round_to_base(4.5, 2) == 4)

    assert(ru.round_to_base(11.5, 20) == 20)
    assert(ru.round_to_base(23.5, 20) == 20)
    assert(ru.round_to_base(34.5, 20) == 40)


# ------------------------------------------------------------------------------
#
def test_round_upper_bound():

    assert(ru.round_upper_bound(0.5) ==  1)
    assert(ru.round_upper_bound(1.5) ==  2)
    assert(ru.round_upper_bound(2.5) ==  5)
    assert(ru.round_upper_bound(4.5) ==  5)
    assert(ru.round_upper_bound(5.5) == 10)
    assert(ru.round_upper_bound(9.5) == 10)

    assert(ru.round_upper_bound( 5000) ==  10000)                   # noqa: E201
    assert(ru.round_upper_bound(15000) ==  20000)
    assert(ru.round_upper_bound(25000) ==  50000)
    assert(ru.round_upper_bound(45000) ==  50000)
    assert(ru.round_upper_bound(55000) == 100000)
    assert(ru.round_upper_bound(95000) == 100000)


# ------------------------------------------------------------------------------
#
def test_sh_callout():

    out, err, ret = ru.sh_callout('echo TRUE')
    assert(out == 'TRUE\n'),  out
    assert(err == ''),        err
    assert(ret == 0),         ret

    out, err, ret = ru.sh_callout('false')
    assert(out == ''),        out
    assert(err == ''),        err
    assert(ret == 1),         ret

    out, err, ret = ru.sh_callout('echo FALSE 1>&2; exit 2', shell=True)
    assert(out == ''),        out
    assert(err == 'FALSE\n'), err
    assert(ret == 2),         ret


# ------------------------------------------------------------------------------
#
def test_sh_callout_async():

    pass

#     import t ime
#     t_0 = time.time()
#     p   = ru.sh_callout_async('echo TRUE && sleep 1', shell=True, stdout=True)
#
#     assert(p.stdout.get() == 'TRUE')
#     assert(p.state        == ru.RUNNING)
#
#     t_1 = time.time()
#
#     assert(p.stdout.get() is None)
#     assert(p.state        == ru.DONE)
#
#     t_2 = time.time()
#
#     assert(t_1 - t_0 < 0.1)
#     assert(t_2 - t_0 > 1.0)


# ------------------------------------------------------------------------------
#
def test_get_env_ns():

    os.environ['RADICAL_UTILS_LOG_LVL'] = 'DEBUG'
    os.environ['RADICAL_LOG_TGT']       = '/dev/null'

    for ns in ['radical.utils.test', 'radical.utils']:

        assert(ru.get_env_ns('LOG_LVL', ns) == 'DEBUG')
        assert(ru.get_env_ns('log.tgt', ns) == '/dev/null')
        assert(ru.get_env_ns('LOG.TGT', ns) == '/dev/null')
        assert(ru.get_env_ns('LOG_TGT', ns) == '/dev/null')
        assert(ru.get_env_ns('TGT_LOG', ns) is None)


# ------------------------------------------------------------------------------
#
def test_expand_env():

    noenv = {'BIZ' : 'biz'}
    env   = {'BAR' : 'bar'}

    os.environ['BAR'] = 'bar'
    os.environ['BIZ'] = 'biz'

    bar = os.environ.get('BAR')
    biz = os.environ.get('BIZ')

    tc = {'${BAR}'             : [bar,                  # os.environ
                                  'bar',                # env
                                  None],                # noenv
          'foo_${BAR}_baz'     : ['foo_%s_baz' % bar,
                                  'foo_bar_baz',
                                  'foo__baz'   ],
          'foo_${BAR:buz}_baz' : ['foo_%s_baz' % bar,
                                  'foo_bar_baz',
                                  'foo_buz_baz'],
          'foo_${BAR:$BIZ}_baz': ['foo_%s_baz' % bar,
                                  'foo_bar_baz',
                                  'foo_%s_baz' % biz],
         }

    # test string expansion (and also create list and dict for other tests
    l = list()
    d = dict()
    i = 0
    for k,v in tc.items():
        assert(ru.expand_env(k       ) == v[0])
        assert(ru.expand_env(k,   env) == v[1])
        assert(ru.expand_env(k, noenv) == v[2])
        l.append(k)
        d[i] = k
        i   += 1

    # test list expansion
    l0 = copy.deepcopy(l)
    l1 = copy.deepcopy(l)
    l2 = copy.deepcopy(l)

    ru.expand_env(l0)
    ru.expand_env(l1, env)
    ru.expand_env(l2, noenv)

    for i,v in enumerate(l):
        assert(l0[i] == tc[v][0])
        assert(l1[i] == tc[v][1])
        assert(l2[i] == tc[v][2])

    # test dict expansion
    d0 = copy.deepcopy(d)
    d1 = copy.deepcopy(d)
    d2 = copy.deepcopy(d)

    ru.expand_env(d0)
    ru.expand_env(d1, env)
    ru.expand_env(d2, noenv)

    for k,v in d0.items(): assert(v == tc[d[k]][0])
    for k,v in d1.items(): assert(v == tc[d[k]][1])
    for k,v in d2.items(): assert(v == tc[d[k]][2])

    # test `ignore_missing` flag
    env = {'BAR' : 'bar'}
    src = 'foo${FIZ}.baz'
    tgt = 'foo.baz'
    assert(ru.expand_env(src, env                     ) == tgt)
    assert(ru.expand_env(src, env, ignore_missing=True) == tgt)

    with pytest.raises(ValueError):
        ru.expand_env(src, env, ignore_missing=False)


# ------------------------------------------------------------------------------
#
def test_script_2_func():

    # create a temp script to convert and run
    [tmpfile, tmpname] = tempfile.mkstemp()
    os.write(tmpfile, ru.as_bytes("""#!/usr/bin/env python3

BUZ = 'buz'

def get_buz():
    return BUZ

if __name__ == '__main__':
    import os,sys
    print('hello')
    sys.stderr.write('world')
    os.system('echo "%%s %%s %%s OK" > %s.out' %% (sys.argv[1], sys.argv[2],
                                                   get_buz()))
    if sys.argv[2] == 'exit':
        exit(2)
    raise ValueError('oops')

""" % tmpname))

    # create a method handle from the tmp script, and call it
    func = ru.script_2_func(tmpname)

    out, err, ret, ec = func('foo bar'.split())
    assert(out == 'hello\n')
    assert(err == 'world')
    assert(ret == 'oops')
    assert(ec  == 1)

    with open(tmpname + '.out', 'r') as fin:
        data = fin.read()
    assert(data.endswith('foo bar buz OK\n')), tmpname

    os.unlink(tmpname + '.out')

    out, err, ret, ec = func('foo', 'exit')
    assert(out == 'hello\n')
    assert(err == 'world')
    assert(ret == 'SystemExit')
    assert(ec  == 2)

    with open(tmpname + '.out', 'r') as fin:
        data = fin.read()
    assert(data.endswith('foo exit buz OK\n')), tmpname

    os.unlink(tmpname)
    os.unlink(tmpname + '.out')


# ------------------------------------------------------------------------------
#
def test_base():

    base = ru.get_base('foo')
    assert(base), base
    assert(base[0] == '/'), base
    base  = base.rstrip('/')
    elems = base.split('/')
    assert(elems[-1] == '.foo'), elems

    base = ru.get_base(ns='foo', module='foo.bar')
    assert(base), base
    assert(base[0] == '/'), base
    base  = base.rstrip('/')
    elems = base.split('/')
    assert(elems[-2] == '.foo'), elems
    assert(elems[-1] == 'bar'), elems

    os.environ['FOO_BASE'] = '/tmp/bar/buz'
    base = ru.get_base('foo')
    assert(base), base
    assert(base[0] == '/'), base
    base  = base.rstrip('/')
    assert(base == '/tmp/bar/buz/.foo'), base


# ------------------------------------------------------------------------------
#
def test_ru_open():

    test_str = 'TEST encoding'
    _, fpath = tempfile.mkstemp()

    with ru.ru_open(fpath, 'w') as fd:
        assert(fd.encoding == 'utf8')
        fd.write(test_str)

    with open(fpath, encoding='utf8') as fd:
        assert(fd.read() == test_str)

    try   : os.remove(fpath)
    except: pass


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_import_file()
    test_round_to_base()
    test_round_upper_bound()
    test_sh_callout()
    test_sh_callout_async()
    test_get_env_ns()
    test_expand_env()
    test_script_2_func()
    test_base()
    test_ru_open()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import pytest

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_dict_mixin():
    """
    Test dict mixin
    """

    # --------------------------------------------------------------------------
    @ru.Lockable
    class Test(ru.DictMixin):

        def __init__(self):
            self._d     = dict()
            self['val'] = 1

        def __getitem__(self, key):
            return self._d.__getitem__(key)

        def __setitem__(self, key, val):
            return self._d.__setitem__(key, val)

        def __delitem__(self, key):
            return self._d.__delitem__(key)

        def keys(self):
            return list(self._d.keys())


    # --------------------------------------------------------------------------
    t = Test()

    assert (t['val']      == 1)
    assert (set(t.keys()) == set(['val']))

    assert ('val'       in t)
    assert ('test1' not in t)
    assert ('test2' not in t)

    t['test1'] =  'test'
    t['test2'] = ['test']

    assert ('val'       in t)
    assert ('test1'     in t)
    assert ('test2'     in t)

    assert (t['val']      == 1       )
    assert (t['test1']    == 'test'  )
    assert (t['test2']    == ['test'])
    assert (set(t.keys()) == set(['val', 'test1', 'test2']))

    del t['test1']

    assert (t['val']      == 1       )
    assert (t['test2']    == ['test'])
    assert (set(t.keys()) == set(['val', 'test2']))

    assert ('val'       in t)
    assert ('test1' not in t)
    assert ('test2'     in t)


# ------------------------------------------------------------------------------
#
def test_dict_merge():

    dict_1 = {'key_shared': 'val_shared_1',
              'key_orig_1': 'val_orig_1'}
    dict_2 = {'key_shared': 'val_shared_2',
              'key_orig_2': 'val_orig_2'}

    try:
        ru.dict_merge(dict_1, dict_2)
        assert (False), 'expected ValueError exception'
    except ValueError:
        pass
    except Exception as e:
        assert (False), 'expected ValueError exception, not %s' % e


    ru.dict_merge(dict_1, dict_2, policy='preserve')

    assert (set(dict_1.keys())   == set(['key_shared', 'key_orig_1',
                                                       'key_orig_2']))
    assert (dict_1['key_shared'] == 'val_shared_1')
    assert (dict_1['key_orig_1'] == 'val_orig_1')
    assert (dict_1['key_orig_2'] == 'val_orig_2')


    ru.dict_merge(dict_1, dict_2, policy='overwrite')

    assert (set(dict_1.keys())   == set(['key_shared', 'key_orig_1',
                                                       'key_orig_2']))
    assert (dict_1['key_shared'] == 'val_shared_2')
    assert (dict_1['key_orig_1'] == 'val_orig_1')
    assert (dict_1['key_orig_2'] == 'val_orig_2')


# ------------------------------------------------------------------------------
#
def test_dict_stringexpand():

    target = {'workdir' : '%(home)s/work/',
              'resource': '%(resource)s'}
    source = {'user'    : 'peer_gynt',
              'protocol': 'ssh',
              'host'    : 'localhost',
              'home'    : '/home/%(user)s',
              'resource': '%(protocol)s://%(host)s/'}

    ru.dict_stringexpand(target, source)

    assert (set(target.keys()) == set(['workdir', 'resource']))
    assert (target['workdir']  == '/home/peer_gynt/work/')
    assert (target['resource'] == 'ssh://localhost/')


# ------------------------------------------------------------------------------
#
def test_dict_diff():

    a  = {'foo': 'bar',
          'baz': {'k1': 1,
                  'k2': [2, -2]}
         }
    b  = {'foo': 'bar',
          'baz': {'k1': 1,
                  'k2': [-2],
                  'k3': 3}
         }
    d1 = ru.dict_diff(a, b)
    d2 = {'baz': {'k2': ['len(1) != len(2)'],
                  'k3': {'a': 3}}}

    assert(d1 == d2)

    with pytest.raises(AssertionError):
        ru.iter_diff(a, b)


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_dict_diff()
    test_dict_mixin()
    test_dict_merge()
    test_dict_stringexpand()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import time
import threading     as mt
import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_lockable():
    '''
    Test lockable decorator
    '''

    # --------------------------------------------------------------------------
    @ru.Lockable
    class Test(object):

        # ----------------------------------------------------------------------
        def __init__(self):

            self.val = False

        # ----------------------------------------------------------------------
        def test(self):

            self.lock()       # lock before spawning thread
            thread_1 = mt.Thread(target=self.test_1)
            thread_1.start()  # thread will run until lock check
            time.sleep(0.1)   # enough time to trigger lock violation
            self.val = False  # set a bogus value
            self.unlock()     # only now thread can set True
            thread_1.join()   # make sure the value was set
            assert(self.val)  # make sure the value was set correctly

            self.lock()       # lock before spawning thread
            thread_2 = mt.Thread(target=self.test_2)
            thread_2.start()  # thread will run until lock check
            time.sleep(0.1)   # enough time to trigger lock violation
            self.val = False  # set a bogus value
            self.unlock()     # only now thread can set True
            thread_2.join()   # make sure the value was set
            assert(self.val)  # make sure the value was set correctly

        # ----------------------------------------------------------------------
        def test_1(self):

            with self:
                self.val = True

        # ----------------------------------------------------------------------
        def test_2(self):

            self.lock()
            self.val = True
            self.unlock()
    # --------------------------------------------------------------------------
    t = Test()

    # check lock with resource manager
    with t:
        pass

    # check explicit and recursive lock/unlock
    t.lock  (); assert(    t.locked())                                    # noqa
    t.unlock(); assert(not t.locked())                                    # noqa
    t.lock  (); assert(    t.locked())                                    # noqa
    t.lock  (); assert(    t.locked())                                    # noqa
    t.unlock(); assert(    t.locked())                                    # noqa
    t.unlock(); assert(not t.locked())                                    # noqa

    # check locking over threads
    t.test()

    # check double unlock
    try                  : t.unlock(); assert(not t.locked())             # noqa
    except RuntimeError  : pass
    except Exception as e: assert(False), "RuntimeError != %s" % type(e)
    else                 : assert(False), "expected RuntimeError, got none"


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_lockable()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

# pylint: disable=no-value-for-parameter,unused-argument,unsubscriptable-object

import radical.utils as ru

from unittest import mock


# ------------------------------------------------------------------------------
#
@mock.patch('radical.utils.zmq.server.Profiler')
def test_zmq_registry(mocked_prof):

    c = None
    r = ru.zmq.Registry()
    r.start()

    try:
        assert(r.addr)
        c = ru.zmq.RegistryClient(url=r.addr)

        c.put('foo.bar.buz', {'biz': 42})
        assert(c.get('foo') == {'bar': {'buz': {'biz': 42}}})
        assert(c.get('foo.bar.buz.biz') == 42)
        assert(c.get('foo.bar.buz.biz.boz') is None)
        assert(c.get('foo') == {'bar': {'buz': {'biz': 42}}})

        assert(c['foo.bar.buz.biz'] == 42)
        assert(c['foo']['bar']['buz']['biz'] == 42)
        assert(c['foo.bar.buz.biz.boz'] is None)

        assert('foo' in c)
        assert(c.keys() == ['foo'])
        del(c['foo'])
        assert('foo' not in c)
        assert(c.keys() == [])

    finally:

        if c:
            c.close()

        r.stop()
        r.wait()


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == '__main__':

    test_zmq_registry()


# ------------------------------------------------------------------------------




__author__    = "Radical.Utils Development Team (Mark Santcroos)"
__copyright__ = "Copyright 2015, RADICAL@Rutgers"
__license__   = "MIT"


import time
import threading
import radical.utils as ru

SIZE = 10
MAX  =  3
ITER = 10


# ------------------------------------------------------------------------------
#
def test_lease_manager():
    """
    Test LeaseManager
    """

    # --------------------------------------------------------------------------
    class Test(object):

        # ----------------------------------------------------------------------
        def __init__(self):

            self.val = False
            self.lm  = None

        # ----------------------------------------------------------------------
        def test(self):

            self.lm = ru.LeaseManager(SIZE)

            iw_thread_1 = threading.Thread(target=self.iw_thread,
                                           kwargs={'uid': 1, 'pool': 'pool1'})
            iw_thread_2 = threading.Thread(target=self.iw_thread,
                                           kwargs={'uid': 2, 'pool': 'pool2'})
            ow_thread_1 = threading.Thread(target=self.ow_thread,
                                           kwargs={'uid': 1, 'pool': 'pool1'})
            ow_thread_2 = threading.Thread(target=self.ow_thread,
                                           kwargs={'uid': 2, 'pool': 'pool2'})
            mon_thread  = threading.Thread(target=self.mon_thread,
                                           kwargs={'uid': 1, 'pool': 'pool1'})
            iw_thread_1.start()
            iw_thread_2.start()
            ow_thread_1.start()
            ow_thread_2.start()
            mon_thread.start()

            iw_thread_1.join()
            iw_thread_2.join()
            ow_thread_1.join()
            ow_thread_2.join()
            mon_thread.join()

        # ----------------------------------------------------------------------
        def iw_thread(self, uid, pool):

            name = "IWo Thread-%d (%s)" % (uid, pool)

            for _ in range(ITER):
                leases = list()
                for _ in range(MAX):
                    lease = self.lm.lease(pool, dict)
                    lease.obj['name'] = name
                    leases.append(lease)

                time.sleep(.1)

                for lease in leases:
                    self.lm.release(lease)

        # ----------------------------------------------------------------------
        def ow_thread(self, uid, pool):

            name = "OWo Thread-%d (%s)" % (uid, pool)

            for _ in range(ITER):
                leases = list()
                for _ in range(MAX):
                    lease = self.lm.lease(pool, dict)
                    lease.obj['name'] = name
                    leases.append(lease)

                time.sleep(.1)

                for lease in leases:
                    self.lm.release(lease)

        # ----------------------------------------------------------------------
        def mon_thread(self, uid, pool):

            name = "Mon Thread-%d (%s)" % (uid, pool)

            for _ in range(ITER):
                leases = list()
                for _ in range(MAX):
                    lease = self.lm.lease(pool, dict)
                    lease.obj['name'] = name
                    leases.append(lease)

                time.sleep(.1)

                for lease in leases:
                    self.lm.release(lease)
        # ----------------------------------------------------------------------

    t = Test()
    t.test()


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_lease_manager()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

# pylint: disable=protected-access

import os


# ------------------------------------------------------------------------------
#
def test_locks():
    '''
    Test debug lock wrappers
    '''

    os.environ['RADICAL_DEBUG']        = 'True'
    os.environ['RADICAL_DEBUG_HELPER'] = 'True'

    import radical.utils as ru

    l  = ru.Lock()
    rl = ru.RLock(name='bar')

    assert(not l.waits)
    assert(not rl.waits)

    with l:
        with rl:
            assert(not l.waits)
            assert(not rl.waits)

    assert(l.name  in ru.debug._debug_helper.locks)                       # noqa
    assert(rl.name in ru.debug._debug_helper.rlocks)                      # noqa

    ru.debug._debug_helper.unregister_lock(l.name)                        # noqa
    ru.debug._debug_helper.unregister_rlock(rl.name)                      # noqa

    assert(l.name  not in ru.debug._debug_helper.locks)                   # noqa
    assert(rl.name not in ru.debug._debug_helper.rlocks)                  # noqa


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_locks()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

# pylint: disable=no-value-for-parameter

import os

from unittest import mock, TestCase

from radical.utils.json_io import write_json

from radical.utils.zmq import Client, Server


# ------------------------------------------------------------------------------
#
class TestZMQClient(TestCase):

    # --------------------------------------------------------------------------
    #
    @mock.patch('radical.utils.zmq.server.Logger')
    @mock.patch('radical.utils.zmq.server.Profiler')
    def test_client(self, mocked_profiler, mocked_logger):

        with self.assertRaises(ValueError):
            # neither `server` nor `url` is set
            Client()

        s = Server()
        s.start()

        write_json({'addr': s.addr}, 'server_local.cfg')
        c = Client(server='server_local')
        self.assertEqual(c.url, s.addr)
        c.close()
        os.unlink('server_local.cfg')

        c = Client(url=s.addr)
        self.assertEqual(c.url, s.addr)

        echo_str = 'test_echo'
        self.assertEqual(c.request(cmd='echo', arg=echo_str), echo_str)

        with self.assertRaises(RuntimeError) as e:
            # unknown attribute for command `echo`
            c.request(cmd='echo', wrong_attr=echo_str)
        self.assertIn('TypeError', str(e.exception))
        self.assertIn('unexpected keyword argument', str(e.exception))

        with self.assertRaises(RuntimeError) as e:
            # no command in request
            c.request(cmd='')
        self.assertIn('no command in request', str(e.exception))

        with self.assertRaises(RuntimeError) as e:
            # command [no_registered_cmd] unknown
            c.request(cmd='no_registered_cmd')
        self.assertIn('command [no_registered_cmd] unknown', str(e.exception))

        c.close()
        s.stop()
        s.wait()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    tc = TestZMQClient()
    tc.test_client()

# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__author__    = 'RADICAL-Cybertools Team'
__copyright__ = 'Copyright 2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import socket

os.environ['RADICAL_BASE'] = '/tmp'

import glob
import shutil

import radical.utils as ru

from unittest import mock, TestCase

TEST_CASES_PATH = 'tests/test_cases/host.*.json'


# ------------------------------------------------------------------------------
#
class HostTestCase(TestCase):

    # --------------------------------------------------------------------------
    #
    @classmethod
    def setUpClass(cls):
        cls._base_dir = ru.get_radical_base('utils')

        cls._test_cases = []
        for f in glob.glob(TEST_CASES_PATH):
            cls._test_cases.extend(ru.read_json(f))

    # --------------------------------------------------------------------------
    #
    @classmethod
    def tearDownClass(cls):
        if hasattr(cls, '_base_dir') and os.path.isdir(cls._base_dir):
            try:
                shutil.rmtree(cls._base_dir)
            except OSError as e:
                print('[ERROR] %s - %s' % (e.filename, e.strerror))

    # --------------------------------------------------------------------------
    #
    @mock.patch('radical.utils.host.socket')
    def test_get_hostname(self, mocked_socket):
        """Check caching wrapper for gethostname."""

        # Note that ru.get_hostname() actually uses getfqdn()
        fq_hostname = socket.getfqdn()
        mocked_socket.gethostname.return_value = socket.gethostname()
        mocked_socket.getfqdn.return_value = fq_hostname

        ru.host._hostname = None                         # pylint: disable=W0212
        self.assertFalse(mocked_socket.gethostname.called)
        self.assertFalse(mocked_socket.getfqdn.called)
        self.assertEqual(ru.get_hostname(), fq_hostname)
        self.assertTrue(mocked_socket.getfqdn.called)

        mocked_socket.getfqdn.reset_mock()
        ru.get_hostname()  # System level hostname lookup is not called 2nd time
        self.assertFalse(mocked_socket.gethostname.called)
        self.assertFalse(mocked_socket.getfqdn.called)

        # reset
        ru.host._hostname = None                         # pylint: disable=W0212

    # --------------------------------------------------------------------------
    #
    def test_create_hostfile(self):

        for test_case in self._test_cases:

            if 'hostlist' not in test_case['input']:
                continue
            hostlist = test_case['input']['hostlist']
            sep      = test_case['input'].get('separator', ' ')

            filename = ru.create_hostfile(self._base_dir, 'tc', hostlist, sep)

            with open(filename, 'r', encoding='utf8') as f:
                hostfile_lines = f.readlines()
            self.assertEqual(test_case['result']['lines'], hostfile_lines)

    # --------------------------------------------------------------------------
    #
    def test_compress_hostlist(self):

        for test_case in self._test_cases:

            if 'hostlist' not in test_case['input']:
                continue
            hostlist = test_case['input']['hostlist']

            self.assertEqual(test_case['result']['hostlist_compressed'],
                             ru.compress_hostlist(hostlist))

    # --------------------------------------------------------------------------
    #
    def test_get_hostlist_by_range(self):

        for test_case in self._test_cases:

            if 'hoststring_range' not in test_case['input']:
                continue
            hoststring = test_case['input']['hoststring_range']

            if test_case['result'] == 'ValueError':

                with self.assertRaises(ValueError):
                    # non numeric set of ranges OR incorrect range format
                    ru.get_hostlist_by_range(hoststring)

            else:

                prefix = test_case['input'].get('prefix', '')
                width  = test_case['input'].get('width', 0)

                self.assertEqual(test_case['result']['hostlist'],
                                 ru.get_hostlist_by_range(hoststring=hoststring,
                                                          prefix=prefix,
                                                          width=width))

    # --------------------------------------------------------------------------
    #
    def test_get_hostlist(self):

        for test_case in self._test_cases:

            if 'hoststring' not in test_case['input']:
                continue
            hoststring = test_case['input']['hoststring']

            if test_case['result'] == 'ValueError':

                with self.assertRaises(ValueError):
                    ru.get_hostlist(hoststring)

            else:

                self.assertEqual(test_case['result']['hostlist'],
                                 ru.get_hostlist(hoststring=hoststring))


# ------------------------------------------------------------------------------



#!/usr/bin/env python

__author__    = "Andre Merzky, Ole Weidner"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import copy

import radical.utils as ru
from   radical.utils.contrib.urlparse25 import urljoin


# ------------------------------------------------------------------------------
#
def test_contrib():

    test_cases = [(None,             ''                    ),
                  ('',               ''                    ),
                  ('http://a/b/c/d', ''                    ),
                  ('g:h',            'g:h'                 ),
                  ('http:g',         'http://a/b/c/g'      ),
                  ('http:',          'http://a/b/c/d'      ),
                  ('g',              'http://a/b/c/g'      ),
                  ('./g',            'http://a/b/c/g'      ),
                  ('g/',             'http://a/b/c/g/'     ),
                  ('/g',             'http://a/g'          ),
                  ('//g',            'http://g'            ),
                  ('?y',             'http://a/b/c/?y'     ),  # [1]
                  ('g?y',            'http://a/b/c/g?y'    ),
                  ('g?y/./x',        'http://a/b/c/g?y/./x'),
                  ('.',              'http://a/b/c/'       ),
                  ('./',             'http://a/b/c/'       ),
                  ('..',             'http://a/b/'         ),
                  ('../',            'http://a/b/'         ),
                  ('../g',           'http://a/b/g'        ),
                  ('../..',          'http://a/'           ),
                  ('../../g',        'http://a/g'          ),
                  ('../../../g',     'http://a/../g'       ),
                  ('./../g',         'http://a/b/g'        ),
                  ('./g/.',          'http://a/b/c/g/'     ),
                  ('/./g',           'http://a/./g'        ),
                  ('g/./h',          'http://a/b/c/g/h'    ),
                  ('g/../h',         'http://a/b/c/h'      ),
                  ('http:g',         'http://a/b/c/g'      ),
                  ('http:',          'http://a/b/c/d'      ),
                  ('http:?y',        'http://a/b/c/?y'     ),  # [1]
                  ('http:g?y',       'http://a/b/c/g?y'    ),
                  ('http:g?y/./x',   'http://a/b/c/g?y/./x')]

    # [1] https://bugs.python.org/issue18828 - open since 2013 :-/
    # This test case *should* result in `http://a/b/c/d?y`


    base = ''
    for url,check in test_cases:

        u      = ru.Url(url)
        result = urljoin(base, url)

        if check:
            assert(result == check), '%s == %s' % (result, check)

        if not base:
            base = result

        if url is None: assert(str(copy.deepcopy(u)) ==     ''), (u, url)
        else          : assert(str(copy.deepcopy(u)) == str(u)), (u, url)

        assert(bool(u) == bool(url)), (str(u), url, bool(u), bool(url))


# -------------------------------------------------------------------------
#
def test_url_api():

    # test basic functionality for valid schemas

    u1 = ru.Url('http://user:pwd@hostname.domain:9999/path?query#fragment')

    assert u1.scheme   == "http"
    assert u1.username == "user"
    assert u1.password == "pwd"
    assert u1.host     == "hostname.domain"
    assert u1.port     == int(9999)
    assert u1.query    == 'query'
    assert u1.fragment == 'fragment'

    u2 = ru.Url()

    u2.scheme   = "http"
    u2.username = "user"
    u2.password = "pwd"
    u2.host     = "hostname.domain"
    u2.port     = int(9999)
    u2.query    = 'query'
    u2.fragment = 'fragment'

    assert u2.scheme   == "http"
    assert u2.username == "user"
    assert u2.password == "pwd"
    assert u2.host     == "hostname.domain"
    assert u2.port     == int(9999)
    assert u2.query    == 'query'
    assert u2.fragment == 'fragment'


# ------------------------------------------------------------------------------
#
def test_url_scheme_issue():

    # test basic functionality for invalid schemas

    u1 = ru.Url("unknownscheme://user:pwd@hostname.domain:9999/path")

    assert u1.scheme   == "unknownscheme"
    assert u1.username == "user"
    assert u1.password == "pwd"
    assert u1.host     == "hostname.domain"
    assert u1.port     == int(9999)


# ------------------------------------------------------------------------------
#
def test_url_issue_49():

    # ensure correct str serialization after setting elements

    url = ru.Url   ("scheme://pass:user@host:123/dir/file?query#fragment")
    url.set_host   ('remote.host.net')
    url.set_scheme ('sftp')
    url.set_path   ('/tmp/data')

    assert str(url) == "sftp://pass:user@remote.host.net:123/tmp/data"


# ------------------------------------------------------------------------------
#
def test_url_issue_61():

    # ensure correct query extraction

    url = ru.Url ("advert://localhost/?dbtype=sqlite3")
    assert url.query == "dbtype=sqlite3"


# ------------------------------------------------------------------------------
#
def test_url_issue_rs_305():
    # This compensates
    #
    #     >>> import os
    #     >>> os.path.normpath('//path//to//dir//')
    #     '//path/to/dir'
    #
    # to a normalization resulting in
    #
    #     '/path/to/dir'
    #
    # as required by the SAGA spec

    url1 = ru.Url ("advert://localhost/path/to/file")
    url2 = ru.Url ("advert://localhost//path/to/file")
    assert url1.path == url2.path


# ------------------------------------------------------------------------------
#
def test_url_properties():

    # test various properties

    url = ru.Url("")
    assert str(url)           == ""

    url.scheme = "scheme"
    assert str(url)           == "scheme://"
    assert url.get_scheme()   == "scheme"

    url.set_scheme("tscheme")
    assert url.get_scheme()   == "tscheme"

    url.scheme = "scheme"
    url.host   = "host"
    assert str(url)           == "scheme://host"
    assert url.get_host()     == "host"

    url.set_host("thost")
    assert url.get_host()     == "thost"

    url.host = "host"
    url.port = None
    assert str(url)           == "scheme://host"
    assert url.get_port()     is None

    url.host = "host"
    url.port = 42
    assert str(url)           == "scheme://host:42"
    assert url.get_port()     == 42

    url.set_port(43)
    assert url.get_port()     == 43

    url.port     = 42
    url.username = "username"
    assert str(url)           == "scheme://username@host:42"
    assert url.get_username() == "username"

    url.set_username("tusername")
    assert url.get_username() == "tusername"

    url.username = "username"
    url.password = "password"
    assert str(url)           == "scheme://username:password@host:42"
    assert url.get_password() == "password"

    url.set_password("tpassword")
    assert url.get_password() == "tpassword"

    url.password = "password"
    url.path     = "/path/"
    assert str(url)           == "scheme://username:password@host:42/path/"
    assert url.get_path()     == "/path/"

    url.set_path("tpath")
    assert url.get_path()     == "/tpath"


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_contrib()
    test_url_api()
    test_url_scheme_issue()
    test_url_issue_49()
    test_url_issue_61()
    test_url_issue_rs_305()
    test_url_properties()


# ------------------------------------------------------------------------------





__author__    = 'Radical.Utils Development Team (Andre Merzky, Ole Weidner)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


# import radical.utils.signatures as rus


# # ----------------------------------------------------------------------------
# @rus.takes   (basestring, int, rus.optional (float))
# @rus.returns (int)
# def sigtest (string, intger, float=3.1415926) :
#     return 1
#
# # ----------------------------------------------------------------------------
# def test_signatures () :
#     '''
#     Test if signature violations are flagged
#     '''
#
#     try                  : ret = sigtest ('string', 2.4, 'hallo')
#     except TypeError as e: pass
#     except Exception as e: assert(0), 'TypeError != %s (%s)' % (type(e), e)
#     else                 : assert(0), 'expected TypeError exception, got none'
#
#     try                  : ret = sigtest ('string', 2, 1.1414)
#     except Exception as e: assert(0), 'exception %s: %s' % (type(e), e)
#
#
# # ----------------------------------------------------------------------------
# # run tests if called directly
# if __name__ == '__main__':
#
#     test_signatures ()
#
#
# # ----------------------------------------------------------------------------




# Copyright 2009 Google Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# Licensed to the PSF under a Contributor Agreement.
#
# Author: Gregory P. Smith <greg@krypto.org>

"""
This module implements a pthread_atfork() work-a-like mechanism for all
fork() calls made from the Python os module.  Any time a fork() is called
from Python a set of unique callbacks will be made in each of the following
three states:
    Preparing to fork - Immediately before the fork call is made.
    In the parent after fork - Immediately after the fork (regardless of
                               success or failure) in the parent process.
    In the child after fork - Immediately after the fork in the child process.

To use this module, first import it early on your programs initialization:

    import atfork
    atfork.monkeypatch_os_fork_functions()

That will stub out os.fork and os.forkpty with wrapped versions implementing
the enhanced behavior.

Next, register your atfork actions by calling atfork.atfork:

    atfork.atfork(prepare=my_lock.acquire,
                  parent=my_lock.release,
                  child=my_lock.release)

No API to unregister an atfork call is provided.  If you are concerned
about resource usage by references your callable holds, consider using
weakref's within your callable.
"""

# flake8: noqa: F401

from .atfork       import monkeypatch_os_fork_functions, atfork
from .stdlib_fixer import fix_logging_module

fix_logging_module()
monkeypatch_os_fork_functions()




#!/usr/bin/env python

__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2016, RADICAL@Rutgers"
__license__   = "MIT"


'''
Unit tests for ru.Heartbeat()
'''

import os
import sys
import time

import multiprocessing as mp

import radical.utils   as ru


# ------------------------------------------------------------------------------
#
def test_hb_default():
    '''
    spawn a process with HB and expect it to live for `tout` seconds.  If it
    dies before, it will return a non-zero exit code.  If it stays alive longer
    than 6 seconds, we consider this an error.
    '''

    dur  = 3.0
    ival = 0.1
    tout = 0.3

    # --------------------------------------------------------------------------
    def get_counter():
        cnt = 0

        def count(test=False):
            nonlocal cnt
            if not test:
                cnt += 1
            return cnt
        return count

    # --------------------------------------------------------------------------
    def proc():

        cb = get_counter()

        def term_cb(uid):

            now = time.time()
            assert(now >= t0 + dur), '%.1f > %.1f + %.1f' % (now, t0, dur)

            cnt_tgt = dur / ival
            cnt_chk = cb(test=True)
            assert(cnt_tgt * 0.8 < cnt_chk < cnt_tgt * 1.2), [cnt_tgt, cnt_chk]

        hb = ru.Heartbeat('foo', timeout=tout, interval=ival,
                                 beat_cb=cb, term_cb=term_cb)
        t0 = time.time()

        hb.start()

        while time.time() < t0 + dur:
            hb.beat()

        while True:
            time.sleep(0.1)
    # --------------------------------------------------------------------------

    p = None
    try:
        p = mp.Process(target=proc)
        p.start()

        # proc should still be alive after 2 seconds
        time.sleep(2)
        assert(p.is_alive())

        # but it should have a zero exit value after 2 more seconds
        time.sleep(2)
        assert(not p.is_alive())
        assert(p.exitcode)

    finally:
        try   : os.kill(p.pid, 9)
        except: pass


# ------------------------------------------------------------------------------
#
def test_hb_uid():
    '''
    Use two different uids for heartbeats, and let only one
    time out.
    '''

    # --------------------------------------------------------------------------
    def proc():

        hb = ru.Heartbeat('test', timeout=0.1, interval=0.01)
        t0 = time.time()

        hb.start()

        try:
            while True:
                if time.time() < t0 + 3: hb.beat('short')
                if time.time() < t0 + 5: hb.beat('long')
                time.sleep(0.05)
            while True:
                time.sleep(1)

        finally:
            if time.time() > t0 + 3.2:
                sys.exit(-1)
    # --------------------------------------------------------------------------

    p = None
    try:
        p = mp.Process(target=proc)
        p.start()

        # proc should still be alive after 2 seconds
        time.sleep(2)
        assert(p.is_alive())

        # but it should have a zero exit value after 2 more seconds
        time.sleep(2)
        assert(not p.is_alive())
        assert(p.exitcode)

    finally:
        try   : os.kill(p.pid, 9)
        except: pass


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_hb_default()
    test_hb_uid()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = 'Radical.Utils Development Team (Andre Merzky, Ole Weidner)'
__copyright__ = 'Copyright 2013, RADICAL@Rutgers'
__license__   = 'MIT'


import resource

import radical.utils as ru


# ------------------------------------------------------------------------------
def test_plugin_manager():
    '''
    Test plugin manager
    '''

    pmgr = ru.PluginManager('radical.utils')
    import pprint
    pprint.pprint(pmgr._registry)

    plugin_2 = pmgr.load('unittests_1', 'default_2', 'a', 1)
    ret = plugin_2.run()
    assert(plugin_2.plugin_type == 'unittests_1')
    assert(plugin_2.plugin_name == 'default_2')
    assert(ret == ('a', 1)), 'plugin_2 invocation: %s != %s' % (['a', 1], ret)

    try:
        pmgr.load('unittests_1', 'default_1')
        assert False
    except LookupError:
        pass
    except:
        assert False


    plugin_2 = pmgr.load('unittests_1', 'default_2', 'a', 1)
    ret = plugin_2.run()
    assert(ret == ('a', 1)), 'plugin_2 invocation: %s != %s' % (['a', 1], ret)

    plugin_3 = pmgr.load('unittests_2', 'default_2', 'a', 1)
    ret = plugin_3.run()
    assert(ret == ('a', 1)), 'plugin_3 invocation: %s != %s' % (['a', 1], ret)

    # load twice -- plugin_2 is marked as singleton plugin_2, and will raise
    # if it is created twice
    plugin_3 = pmgr.load('unittests_2', 'default_2', 'a', 1)
    ret = plugin_3.run()
    assert(ret == ('a', 1)), 'plugin_3 invocation: %s != %s' % (['a', 1], ret)


    mem_0 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    pmgr  = ru.PluginManager('radical.utils')

    for i in range(1000000):
        plugin_4 = pmgr.load('unittests_2', 'default_2', 'a', 1)
        ret = plugin_4.run()
        assert(ret == ('a', 1)), 'plugin_4 invoc: %s != %s' % (['a', 1], ret)

        if not i % 100000:
            mem_1 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
            assert(mem_1 <= 2 * mem_0)

    mem_1 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    assert(mem_1 <= 2 * mem_0)


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == '__main__':

    test_plugin_manager()


# ------------------------------------------------------------------------------






__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"


# ------------------------------------------------------------------------------
#
"""
Using the RU logging module can lead to deadlocks when used in multiprocess and
multithreaded environments.  This is due to the native python logging library
not being thread save.  See [1] for a Python bug report on the topic from 2009.
As of 2016, a patch is apparently submitted, but not yet accepted, for Python
3.5.  So we have to solve the problem on our own.

The fix is encoded in the 'after_fork()' method.  That method *must* be called
immediately after a process fork.  A fork will carry over all locks from the
parent process -- but it will not carry over any threads.  That means that, if
a fork happens while a thread owns the lock, the child process will start with
a locked lock -- but since there are no threads, there is actually nobody who
can unlock the lock, thus badaboom.

Since after the fork we *know* the logging locks should be unset (after all, we
are not in a logging call right now, are we?), we pre-emtively unlock them here.
But, to do so we need to know what locks exist in the first place.  For that
purpose, we create a process-singleton of all the loggers we hand out via
'get_logger()'.

Logging locks are 'threading.RLock' instances.  As such they can be locked
multiple times (from within the same thread), and we have to unlock them that
many times.  We use a shortcut, and create a new, unlocked lock.
"""


# ------------------------------------------------------------------------------
#
import os
import sys
import threading
import colorama
import logging

from   .atfork    import atfork
from   .misc      import get_env_ns       as ru_get_env_ns
from   .debug     import import_module    as ru_import_module
from   .config    import DefaultConfig
from   .singleton import Singleton


CRITICAL = logging.CRITICAL
ERROR    = logging.ERROR
WARNING  = logging.WARNING
WARN     = logging.WARNING
INFO     = logging.INFO
DEBUG    = logging.DEBUG
OFF      = -1


# ------------------------------------------------------------------------------
#
class _LoggerRegistry(object, metaclass=Singleton):

    from .singleton import Singleton

    def __init__(self):
        self._registry = list()

    def add(self, logger):
        self._registry.append(logger)

    def release_all(self):
        for logger in self._registry:
            while logger:
                for handler in logger.handlers:
                    handler.lock = threading.RLock()
                  # handler.reset()
                logger = logger.parent

    def close_all(self):

        for logger in self._registry:
            while logger:
                for handler in logger.handlers:
                    handler.close()
                    logger.removeHandler(handler)
                logger = logger.parent
        self._registry = list()


# ------------------------------------------------------------------------------
#
_logger_registry = _LoggerRegistry()


# ------------------------------------------------------------------------------
def _after_fork():

    _logger_registry.release_all()
    logging._lock = threading.RLock()         # pylint: disable=protected-access


# ------------------------------------------------------------------------------
#
def _atfork_prepare():
    pass


# ------------------------------------------------------------------------------
#
def _atfork_parent():
    pass


# ------------------------------------------------------------------------------
#
def _atfork_child():
    _after_fork()


# ------------------------------------------------------------------------------
#
atfork(_atfork_prepare, _atfork_parent, _atfork_child)


# ------------------------------------------------------------------------------
#
class ColorStreamHandler(logging.StreamHandler):
    """
    A colorized output SteamHandler
    """

    # --------------------------------------------------------------------------
    #
    def __init__(self, target):

        logging.StreamHandler.__init__(self, target)
        self._tty    = self.stream.isatty()
        self._term   = getattr(self, 'terminator', '\n')
        self.colours = {'DEBUG'    : colorama.Fore.CYAN,
                        'INFO'     : colorama.Fore.GREEN,
                        'WARN'     : colorama.Fore.YELLOW,
                        'WARNING'  : colorama.Fore.YELLOW,
                        'ERROR'    : colorama.Fore.RED,
                        'CRITICAL' : colorama.Back.RED + colorama.Fore.WHITE,
                        'RESET'    : colorama.Style.RESET_ALL + self._term
                       }


    # --------------------------------------------------------------------------
    #
    def emit(self, record):

        # only write in color when using a tty
        if self._tty:
            self.stream.write('%s%s%s' % (self.colours[record.levelname],
                                          self.format(record),
                                          self.colours['RESET']))
        else:
            self.stream.write(self.format(record) + self._term)
        self.stream.flush()


# ------------------------------------------------------------------------------
#
class FSHandler(logging.FileHandler):

    def __init__(self, target):

        try:
            os.makedirs(os.path.abspath(os.path.dirname(target)))
        except:
            pass  # exists

        logging.FileHandler.__init__(self, target)


# ------------------------------------------------------------------------------
#
class Logger(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, name, ns=None, path=None, targets=None, level=None,
                 verbose=False):

        ru_def = DefaultConfig()

        """
        Get a logging handle.

        `name`    is used to identify log entries on this handle.
        `ns`      is used as environment name space to check for the log level
        `targets` is a comma separated list (or Python list) of specifiers,
                  where specifiers are:

                    `0`      : /dev/null
                    `null`   : /dev/null
                    `-`      : stdout
                    `1`      : stdout
                    `stdout` : stdout
                    `=`      : stderr
                    `2`      : stderr
                    `stderr` : stderr
                    `.`      : logfile named ./<name>.log
                    <string> : logfile named <string>

        `path`    file system location to write logfiles to (created as needed)
        `level`   log level (DEBUG, INFO, WARNING, ERROR, CRITICAL, OFF)

        If `ns` is, for example, set to `radical.utils`, then the following
        environment variables are evaluated:

            RADICAL_UTILS_LOG_LVL
            RADICAL_LOG_LVL

            RADICAL_UTILS_LOG_TGT
            RADICAL_LOG_TGT

        The first found variable of each pair is then used for the respective
        settings.
        """

        self._logger = logging.getLogger(name)
        self._logger.propagate = False   # let messages not trickle upward
        self._logger.name      = name

        # otherwise configure this logger
        if not path:
            path = ru_def['log_dir']

        if not ns:
            ns = name

        if not targets:
            targets = ru_get_env_ns('log_tgt', ns)
            if not targets:
                targets = ru_def['log_tgt']

        if isinstance(targets, str):
            targets = targets.split(',')

        if not isinstance(targets, list):
            targets = [targets]

        if not level:
            level = ru_get_env_ns('log_lvl', ns)

        if not level:
            # backward compatibility
            level = ru_get_env_ns('verbose', ns)

        if not level:
            level = ru_def['log_lvl']

        if level in [OFF, 'OFF']:
            targets = ['null']

        try:
            level = int(level)
        except:
            pass

        self._debug_level = 0
        if isinstance(level, int):
            if level < 10:
                self._debug_level = 10 - level
                level = 'DEBUG'

        elif level.upper().startswith('DEBUG_'):
            self._debug_level = int(level.split('_', 1)[1])
            level = 'DEBUG'


        # translate numeric levels into upper case symbolic ones
        levels  = {'50' : 'CRITICAL',
                   '40' : 'ERROR',
                   '30' : 'WARNING',
                   '20' : 'INFO',
                   '10' : 'DEBUG',
                    '0' :  ru_def['log_lvl'],
                   '-1' : 'OFF'}


        level   = levels.get(str(level), str(level)).upper()
        warning = None
        if level not in list(levels.values()):
            warning = "invalid loglevel '%s', use '%s'" \
                                      % (level, ru_def['log_lvl'])
            level   = ru_def['log_lvl']

        formatter = logging.Formatter('%(created).3f : '
                                      '%(name)-20s : '
                                      '%(process)-5d : '
                                      '%(thread)-5d : '
                                      '%(levelname)-8s : '
                                      '%(message)s')

      # print('%-30s -> %-10s %d' % (name, level, self._debug_level))

        # add a handler for each targets (using the same format)
        if not self._logger.handlers:
            p = path
            n = name
            for t in targets:
                if   t in ['0', 'null']       : h = logging.NullHandler()
                elif t in ['-', '1', 'stdout']: h = ColorStreamHandler(sys.stdout)
                elif t in ['=', '2', 'stderr']: h = ColorStreamHandler(sys.stderr)
                elif t in ['.']               : h = FSHandler("%s/%s.log" % (p, n))
                elif t.startswith('/')        : h = FSHandler(t)
                else                          : h = FSHandler("%s/%s"     % (p, t))

                h.setFormatter(formatter)
                h.name = self._logger.name
                self._logger.addHandler(h)

            if level != 'OFF':
                self._logger.setLevel(level)

            if warning:
                self._logger.warning(warning)

        # if `name` points to module, try to log its version info
        if verbose:
            try:
                self._logger.info("%-20s version: %s", 'python.interpreter',
                                  ' '.join(sys.version.split()))

                mod = ru_import_module(name)
                if hasattr(mod, 'version_detail'):
                    self._logger.info("%-20s version: %s",
                                      name, getattr(mod, 'version_detail'))

                elif hasattr(mod, 'version'):
                    self._logger.info("%-20s version: %s",
                                      name, getattr(mod, 'version'))
            except:
                pass

            # also log pid and tid
            try:
                self._logger.info("%-20s pid/tid: %s/%s", '', os.getpid(),
                            threading.current_thread().name)
            except:
                pass

        # keep the handle a round, for cleaning up on fork
        _logger_registry.add(self._logger)

        # store properties
        self._name    = name
        self._ns      = ns
        self._path    = path
        self._level   = level
        self._targets = targets

        # backward compatibility
        self._logger.warn = self._logger.warning


    # --------------------------------------------------------------------------
    #
    @property
    def name(self):
        return self._name

    @property
    def ns(self):
        return self._ns

    @property
    def path(self):
        return self._path

    @property
    def level(self):
        return self._level

    @property
    def targets(self):
        return self._targets


    # --------------------------------------------------------------------------
    #
    # treat `self.debug_1()`, `self.debug_2()` etc. the same as `self.debug()`
    # if the respective `debug_level` is set, and ignore otherwise.  All other
    # unknown method calls are forwarded to the nativ logger instance.  This is
    # basically inheritance, but since the logger class has no constructor, we
    # do it this way.
    #
    def __getattr__(self, name):

        if name.startswith('debug_'):
            level = int(name.split('_', 1)[1])
            if level > self._debug_level:
                return self._ignore
            else:
                return self.debug
        else:
            return getattr(self._logger, name)


    # --------------------------------------------------------------------------
    #
    def _ignore(self, *args, **kwargs):
        pass


    # --------------------------------------------------------------------------
    #
    # Add a close method to make sure we can close file handles etc.  This also
    # closes handles on all parent loggers (if those exist).
    #
    def close(self):

        logger = self._logger
        while logger:
            for handler in logger.handlers:
                handler.close()
                logger.removeHandler(handler)
            logger = logger.parent


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = 'Radical.Utils Development Team'
__copyright__ = 'Copyright 2019, RADICAL@Rutgers'
__license__   = 'MIT'


import time
import threading     as mt

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_zmq_pubsub():
    '''
    create a bridge, 2 producers (A, B) and 2 consumers (C, D).  Send with the
    following rates for 10 seconds:

      A: 10/s
      B: 20/s

    Ensure that
      - the ratios of sent / received messages reflects the rates
      - the local order of messages is preserved
      - messages are received exactly once (no messages get lost / duplicated)
    '''

    c_a = 200
    c_b = 400

    cfg = ru.Config(cfg={'uid'      : 'test_pubsub',
                         'channel'  : 'test',
                         'kind'     : 'pubsub',
                         'log_level': 'error',
                         'path'     : '/tmp/',
                         'sid'      : 'test_sid',
                         'bulk_size': 0,
                         'stall_hwm': 1,
                        })

    b = ru.zmq.PubSub(cfg)
    b.start()

    assert(b.addr_in  != b.addr_out)
    assert(b.addr_in  == b.addr_pub)
    assert(b.addr_out == b.addr_sub)

    data = dict()
    for i in 'ABCD':
        data[i] = dict()
        for j in 'AB':
            data[i][j] = 0

    def cb(uid, topic, msg):
        if msg['idx'] is None:
            return False
        data[uid][msg['src']] += 1

    cb_C = lambda t,m: cb('C', t, m)
    cb_D = lambda t,m: cb('D', t, m)

    ru.zmq.Subscriber(channel=cfg['channel'], url=str(b.addr_sub),
                      topic='topic', cb=cb_C)
    ru.zmq.Subscriber(channel=cfg['channel'], url=str(b.addr_sub),
                      topic='topic', cb=cb_D)
    time.sleep(0.1)

    # --------------------------------------------------------------------------
    def work_pub(uid, n, delay):

        pub = ru.zmq.Publisher(channel=cfg['channel'], url=str(b.addr_pub))
        idx = 0
        while idx < n:
            time.sleep(delay)
            pub.put('topic', {'src': uid,
                              'idx': idx})
            idx += 1
            data[uid][uid] += 1

        # send EOF
        pub.put('topic', {'src': uid,
                          'idx': None})
    # --------------------------------------------------------------------------


    t_a = mt.Thread(target=work_pub, args=['A', c_a, 0.005])
    t_b = mt.Thread(target=work_pub, args=['B', c_b, 0.005])

    t_a.start()
    t_b.start()

    t_a.join()
    t_b.join()

    b.stop()
    time.sleep(0.1)

    assert(data['A']['A'] == c_a)
    assert(data['B']['B'] == c_b)

    assert(data['C']['A'] + data['C']['B'] +
           data['D']['A'] + data['D']['B'] == 2 * (c_a + c_b))


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == '__main__':

    test_zmq_pubsub()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__author__    = 'RADICAL-Cybertools Team'
__copyright__ = 'Copyright 2012-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
os.environ['RADICAL_BASE'] = '/tmp'

import shutil

import radical.utils as ru

from unittest import TestCase


# ------------------------------------------------------------------------------
#
class LoggerTestCase(TestCase):

    # --------------------------------------------------------------------------
    #
    @classmethod
    def setUpClass(cls):
        cls._base_dir = ru.get_radical_base('utils')

    # --------------------------------------------------------------------------
    #
    @classmethod
    def tearDownClass(cls):
        if hasattr(cls, '_base_dir') and os.path.isdir(cls._base_dir):
            try:
                shutil.rmtree(cls._base_dir)
            except OSError as e:
                print('[ERROR] %s - %s' % (e.filename, e.strerror))

    # --------------------------------------------------------------------------
    #
    def test_init(self):

        logger_name = 'log_name_default'
        l = ru.Logger(logger_name)
        log_path = os.path.join(os.getcwd(), '%s.log' % logger_name)
        self.assertTrue(os.path.isfile(log_path))

        self.assertEqual(l.name,    logger_name)
        self.assertEqual(l.ns,      logger_name)
        # default values
        self.assertEqual(l.path,    os.getcwd())
        self.assertEqual(l.level,   'ERROR')
        self.assertEqual(l.targets, ['.'])

        os.remove(log_path)
        self.assertFalse(os.path.isfile(log_path))
        l.close()

        # set log directory as an input parameter `path`
        l = ru.Logger(name='log_base', path=self._base_dir)
        log_path = os.path.join(self._base_dir, 'log_base.log')
        self.assertTrue(os.path.isfile(log_path))
        l.close()

        # set log file using a corresponding environment variable
        log_env_tgt_saved = os.environ.get('RADICAL_UTILS_LOG_TGT', -1)
        log_path = os.path.join(self._base_dir, 'log_base_tgt.log')
        os.environ['RADICAL_UTILS_LOG_TGT'] = log_path
        l = ru.Logger('log_base_tgt', ns='radical.utils')
        self.assertTrue(os.path.isfile(log_path))
        l.close()
        if log_env_tgt_saved == -1:
            del os.environ['RADICAL_UTILS_LOG_TGT']
        else:
            os.environ['RADICAL_UTILS_LOG_TGT'] = log_env_tgt_saved

    # --------------------------------------------------------------------------
    #
    def test_log_levels(self):

        log_path = os.path.join(self._base_dir, 'log_lvl.log')
        l = ru.Logger('log_lvl', targets=[log_path])

        l.setLevel('INFO')

        l.debug('debug')  # the one record that will not be recorded
        l.info('info')
        l.warning('warning')
        l.error('error')
        l.fatal('fatal')
        l.critical('critical')

        with open(log_path) as fd:
            log_records = fd.readlines()

        debug_found, non_debug_found = False, False
        for log_record in log_records:
            if 'debug' in log_record:
                debug_found = True
            if         'info'     in log_record \
                    or 'warning'  in log_record \
                    or 'error'    in log_record \
                    or 'fatal'    in log_record \
                    or 'critical' in log_record:
                non_debug_found = True
            else:
                non_debug_found = False

        self.assertFalse(debug_found)
        self.assertTrue(non_debug_found)
        l.close()

    # --------------------------------------------------------------------------
    #
    def test_debug_level(self):

        log_path = os.path.join(self._base_dir, 'log_debug_lvl.log')
        l = ru.Logger('log_debug_lvl', targets=[log_path], level='DEBUG_2')

        l.debug_1('debug_1')
        l.debug_2('debug_2')
        l.debug_3('debug_3')

        with open(log_path) as fd:
            log_records = fd.readlines()

        debug_3_found = False
        for log_record in log_records:
            self.assertTrue('debug_' in log_record)
            if 'debug_3' in log_record:
                debug_3_found = True

        self.assertFalse(debug_3_found)
        l.close()

# ------------------------------------------------------------------------------


if __name__ == '__main__':

    tc = LoggerTestCase()
    tc.test_init()
    tc.test_log_levels()
    tc.test_debug_level()


# ------------------------------------------------------------------------------



#!/usr/bin/env python

__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"


import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_re_string():
    '''
    Test regex matching
    '''

    txt   = ru.ReString ("The quick brown fox jumps over the lazy dog")
    tgt_l = [' qu', 'ick brown fox jumps']
    tgt_d = {'x'  : 'ick brown fox jumps'}

    with txt // r'(\s.u)(?P<x>.*?j\S+)' as res:
        assert (res)
        assert (len(res) == len(tgt_l))
        for a,b in zip(res,tgt_l):
            assert(a == b)
        assert (res      == tgt_l), "%s != %s" % (str(res), str(tgt_l))
        assert (res[0]   == tgt_l[0])
        assert (res[1]   == tgt_l[1])
        assert (res['x'] == tgt_d['x'])
        assert (res.x    == tgt_d['x'])

        for i, r in enumerate (res):
            assert (r    == tgt_l[i])

    if  txt // '(rabbit)':
        assert (False)

    elif  txt // r'((?:\s).{12,15}?(\S+))':
        assert (True)

    else:
        assert (False)


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_re_string ()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

# pylint: disable=no-value-for-parameter, protected-access

from typing   import Any, List, Dict
from unittest import mock, TestCase

import radical.utils as ru

from radical.utils.zmq.server import zmq
from radical.utils.zmq        import Server


# --------------------------------------------------------------------------
#
class MyZMQServer(ru.zmq.Server):

    def __init__(self):

        ru.zmq.Server.__init__(self)

        self.register_request('test_0', self._test_0)
        self.register_request('test_1', self._test_1)
        self.register_request('test_2', self._test_2)
        self.register_request('test_3', self._test_3)
        self.register_request('test_4', self._test_4)


    def _test_0(self) -> str:
        return 'default'


    def _test_1(self, foo: Any = None) -> Any:
        return foo


    def _test_2(self, foo: Any,
                      bar: Any) -> List[Any]:
        return [foo, bar]


    def _test_3(self, foo: Any,
                      bar: Any = 'default') -> Dict[str, Any]:
        return {'foo': foo,
                'bar': bar}


    def _test_4(self, *args, **kwargs) -> List[Any]:
        return list(args) + list(kwargs.values())


# ------------------------------------------------------------------------------
#
class TestZMQServer(TestCase):

    # --------------------------------------------------------------------------
    #
    @mock.patch('radical.utils.zmq.server.Logger')
    @mock.patch('radical.utils.zmq.server.Profiler')
    def test_init(self, mocked_profiler, mocked_logger):

        s = Server()
        self.assertTrue  (s.uid.startswith('server'))
        self.assertIsNone(s.addr)
        self.assertEqual (s._url, 'tcp://*:10000-11000')

        self.assertFalse(s._up.is_set())
        self.assertFalse(s._term.is_set())

        uid = 'test.server'
        s = Server(uid=uid)
        self.assertEqual(s.uid, uid)

        with self.assertRaises(AssertionError):
            # port(s) not set
            Server(url='tcp://*')

        with self.assertRaises(RuntimeError):
            # port(s) set incorrectly
            Server(url='tcp://*:10000-11000-22000')

        # default callbacks
        self.assertIn('echo', s._cbs)
        self.assertIn('fail', s._cbs)
        self.assertEqual(s._cbs['echo']('test_echo'), 'test_echo')
        with self.assertRaises(RuntimeError):
            s._cbs['fail'](None)

    # --------------------------------------------------------------------------
    #
    @mock.patch.object(Server, '__init__', return_value=None)
    @mock.patch('radical.utils.zmq.server.Logger')
    @mock.patch('radical.utils.zmq.server.Profiler')
    def test_exec_output(self, mocked_profiler, mocked_logger, mocked_init):

        s = Server()

        response_msg = 'response-00'
        output = s._success(res=response_msg)
        self.assertIsInstance(output, dict)
        self.assertIsNone(output['err'])
        self.assertIsNone(output['exc'])
        self.assertEqual (output['res'], response_msg)

        error_msg, exception_msg = 'error-00', 'exception-00'
        output = s._error(err=error_msg, exc=exception_msg)
        self.assertEqual (output['err'], error_msg)
        self.assertEqual (output['exc'], exception_msg)
        self.assertIsNone(output['res'])

        output = s._error()
        self.assertEqual(output['err'], 'invalid request')

    # --------------------------------------------------------------------------
    #
    @mock.patch('radical.utils.zmq.server.Logger')
    @mock.patch('radical.utils.zmq.server.Profiler')
    def test_start(self, mocked_profiler, mocked_logger):

        s = Server(url='tcp://*:12345')
        s.start()
        self.assertTrue(s.addr.endswith('12345'))

        self.assertTrue(s._up.is_set())

        with self.assertRaises(RuntimeError):
            # `start()` can be called only once
            s.start()

        s2 = Server(url='tcp://*:12345-')
        s2.start()
        # logged warning about port already in use
        self.assertTrue(s2._log.warn.called)
        self.assertTrue(s2.addr.endswith('12346'))

        s2.stop()
        s2.wait()

        s.stop()
        s.wait()

        self.assertTrue(s._term.is_set())

    # --------------------------------------------------------------------------
    #
    @mock.patch('radical.utils.zmq.server.zmq.Context')
    @mock.patch('radical.utils.zmq.server.Logger')
    @mock.patch('radical.utils.zmq.server.Profiler')
    def test_zmq(self, mocked_profiler, mocked_logger, mocked_zmq_ctx):

        s = Server()
        mocked_zmq_ctx().socket().bind = mock.Mock(
            side_effect=zmq.error.ZMQError(msg='random ZMQ error'))

        with self.assertRaises(zmq.error.ZMQError):
            s._work()


    # --------------------------------------------------------------------------
    #
    @mock.patch('radical.utils.zmq.server.Logger')
    @mock.patch('radical.utils.zmq.server.Profiler')
    def test_server_class(self, mocked_profiler, mocked_logger):

        s = MyZMQServer()

        try:
            s.start()
            self.assertIsNotNone(s.addr)

            c = ru.zmq.Client(url=s.addr)
            self.assertEqual(c.url, s.addr)

            with self.assertRaisesRegex(RuntimeError, 'no command'):
                c.request('')

            with self.assertRaisesRegex(RuntimeError, 'command .* unknown'):
                c.request('no_registered_cmd')

            with self.assertRaisesRegex(RuntimeError,
                    '.* _test_0.* takes 1 positional argument'):
                c.request('test_0', None)

            ret = c.request('test_0')
            self.assertIsInstance(ret, str)
            self.assertEqual(ret, 'default')

            ret = c.request('test_1', 'foo')
            self.assertIsInstance(ret, str)
            self.assertEqual(ret, 'foo')

            ret = c.request('test_1', ['foo', 'bar'])
            self.assertIsInstance(ret, list)
            self.assertEqual(ret, ['foo', 'bar'])

            ret = c.request('test_1')
            self.assertEqual(ret, None)

            ret = c.request('test_2', 'foo', 'bar')
            self.assertIsInstance(ret, list)
            self.assertEqual(ret, ['foo', 'bar'])

            ret = c.request('test_3', 'foo')
            self.assertIsInstance(ret, dict)
            self.assertEqual(ret, {'foo': 'foo', 'bar': 'default'})

            ret = c.request('test_3', foo='foo', bar='bar')
            self.assertIsInstance(ret, dict)
            self.assertEqual(ret, {'foo': 'foo', 'bar': 'bar'})

            ret = c.request('test_3', 'foo', bar='bar')
            self.assertIsInstance(ret, dict)
            self.assertEqual(ret, {'foo': 'foo', 'bar': 'bar'})

            ret = c.request('test_3', 'foo', 'bar')
            self.assertIsInstance(ret, dict)
            self.assertEqual(ret, {'foo': 'foo', 'bar': 'bar'})

            ret = c.request('test_4', 'foo', 'bar')
            self.assertIsInstance(ret, list)
            self.assertEqual(ret, ['foo', 'bar'])

            ret = c.request('test_4', 'foo', ['bar'])
            self.assertIsInstance(ret, list)
            self.assertEqual(ret, ['foo', ['bar']])

            c.close()

        finally:
            s.stop()
            s.wait()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    tc = TestZMQServer()
    tc.test_init()
    tc.test_exec_output()
    tc.test_start()
    tc.test_zmq()
    tc.test_server_class()

# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = 'RADICAL-Cybertools Team'
__copyright__ = 'Copyright 2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import shutil

import radical.utils as ru

from unittest import TestCase


# ------------------------------------------------------------------------------
#
class StackTestClass(TestCase):

    # --------------------------------------------------------------------------
    #
    @classmethod
    def setUpClass(cls):
        ns_mod              = ru.import_module('radical')
        cls._radical_ns_dir = ns_mod.__path__._path[0]
        # create a directory inside "radical" namespace (false package)
        ru.rec_makedir(cls._radical_ns_dir + '/dummy')

    # --------------------------------------------------------------------------
    #
    @classmethod
    def tearDownClass(cls):
        # delete false package
        try:
            shutil.rmtree(cls._radical_ns_dir + '/dummy')
        except OSError as e:
            print('[ERROR] %s - %s' % (e.filename, e.strerror))

    # --------------------------------------------------------------------------
    #
    def test_stack(self):

        # - "radical" namespace (default) -
        stack_output = ru.stack()
        self.assertIn('sys', stack_output)
        self.assertIn('radical', stack_output)
        for module in ['pilot', 'saga', 'utils']:
            self.assertIn('radical.%s' % module, stack_output['radical'])
        self.assertEqual(stack_output['radical']['radical.dummy'], '?')

        os.environ['RADICAL_DEBUG'] = 'TRUE'
        stack_output = ru.stack()
        self.assertTrue(stack_output['radical']['radical.dummy'].endswith(
            "no attribute 'version_detail'"))

        # - "requests" - not a namespace, but a package-
        stack_output = ru.stack(ns=['radical', 'requests'])
        self.assertFalse(stack_output['requests'])

        # - non existed namespace/package -
        with self.assertRaises(ModuleNotFoundError):
            _ = ru.stack(ns='no_valid_pkg')

# ------------------------------------------------------------------------------




import zmq
import msgpack

from typing import Any

import threading as mt

from ..json_io import read_json

from .utils    import no_intr, sock_connect


# ------------------------------------------------------------------------------
#
_LINGER_TIMEOUT    = 1000  # ms to linger after close
_HIGH_WATER_MARK   = 1024  # number of messages to buffer before dropping
_DEFAULT_BULK_SIZE =    1  # number of messages to put in a bulk


# ------------------------------------------------------------------------------
#
class Client(object):

    # --------------------------------------------------------------------------
    #
    def __init__(self, server: str = None,
                       url:    str = None) -> None:

        if server:
            self._url = read_json('%s.cfg' % server)['addr']

        elif url:
            self._url = url

        else:
            raise ValueError('need server name/cfg or Url')

        self._cb   = None

        self._ctx  = zmq.Context()
        self._sock = self._ctx.socket(zmq.REQ)

        self._sock.linger = _LINGER_TIMEOUT
        self._sock.hwm    = _HIGH_WATER_MARK

        sock_connect(self._sock, self._url)

        self._term   = mt.Event()
        self._active = False


    # --------------------------------------------------------------------------
    #
    @property
    def url(self) -> str:
        return self._url


    # --------------------------------------------------------------------------
    #
    def request(self, cmd: str, *args: Any, **kwargs: Any) -> Any:

        req = msgpack.packb({'cmd'   : cmd,
                             'args'  : args,
                             'kwargs': kwargs})

        no_intr(self._sock.send, req)

        msg = no_intr(self._sock.recv)
        res = msgpack.unpackb(msg)

        if res['err']:
            err_msg = 'ERROR: %s' % res['err']
            if res['exc']:
                err_msg += '\n%s' % ''.join(res['exc'])
            raise RuntimeError(err_msg)

        return res['res']


    # --------------------------------------------------------------------------
    #
    def close(self) -> None:

        self._sock.close()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = 'Radical.Utils Development Team'
__copyright__ = 'Copyright 2019, RADICAL@Rutgers'
__license__   = 'MIT'


import time
import pytest
import threading     as mt

import radical.utils as ru


# ------------------------------------------------------------------------------
#
@pytest.mark.skip(reason="test has a timing problem and frequently fails")
def test_zmq_queue():
    '''
    create a bridge, 2 producers (A, B) and 2 consumers (C, D).  Send with the
    following rates for 10 seconds:

      A: 10/s
      B: 20/s

    Ensure that
      - the ratios of sent / received messages reflects the rates
      - the local order of messages is preserved
      - messages are received exactly once (no messages get lost / duplicated)
    '''

    c_a = 100
    c_b = 200

    cfg = ru.Config(cfg={'uid'      : 'test_queue',
                         'channel'  : 'test',
                         'kind'     : 'queue',
                         'log_level': 'error',
                         'path'     : '/tmp/',
                         'sid'      : 'test_sid',
                         'bulk_size': 50,
                         'stall_hwm': 1,
                        })

    b = ru.zmq.Queue(cfg)
    b.start()

    assert(b.addr_in  != b.addr_out)
    assert(b.addr_in  == b.addr_put)
    assert(b.addr_out == b.addr_get)

    C = ru.zmq.Getter(channel=cfg['channel'], url=str(b.addr_get))
    D = ru.zmq.Getter(channel=cfg['channel'], url=str(b.addr_get))

    A = ru.zmq.Putter(channel=cfg['channel'], url=str(b.addr_put))
    B = ru.zmq.Putter(channel=cfg['channel'], url=str(b.addr_put))

    data = dict()

    def work_put(putter, uid, n, delay):

        data[uid] = list()
        idx   = 0
        while idx < n:
            time.sleep(delay)
            putter.put({'src' : uid,
                        'idx' : idx})
            idx += 1
            data[uid].append(uid)

        # send EOF
        putter.put({'src' : uid,
                    'idx' : None})

    def work_get(getter, uid):

        data[uid] = list()
        done      = False
        n         = 0
        while not done:
            msgs = getter.get()
            for msg in msgs:
                msg = ru.as_string(msg)
                if msg['idx'] is None:
                    done = True
                else:
                    data[uid].append(msg['src'])
                    n += 1

        getter.stop()

    t_a = mt.Thread(target=work_put, args=[A, 'A', c_a, 0.02])
    t_b = mt.Thread(target=work_put, args=[B, 'B', c_b, 0.01])
    t_c = mt.Thread(target=work_get, args=[C, 'C'])
    t_d = mt.Thread(target=work_get, args=[D, 'D'])

    t_a.daemon = True
    t_b.daemon = True
    t_c.daemon = True
    t_d.daemon = True

    t_a.start()
    t_b.start()
    t_c.start()
    t_d.start()

    time.sleep(3)
    b.stop()

  # uids = list(data.keys())
  # for x in uids:
  #     for y in uids:
  #         print('%s: %s: %d' % (x, y, data[x].count(y)))
  #
  # print(len(data['A']))
  # print(len(data['B']))
  # print(len(data['C']))
  # print(len(data['D']))

    assert(data['A'].count('A') == c_a)
    assert(data['B'].count('B') == c_b)
    assert(len(data['A'])       == c_a)
    assert(len(data['B'])       == c_b)

    assert(data['C'].count('A') + data['C'].count('B') +
           data['D'].count('A') + data['D'].count('B') == c_a + c_b)

    avg = (c_a + c_b) / 2
    assert(avg - 30 < data['C'].count('A') + data['C'].count('B') < avg + 30)
    assert(avg - 30 < data['D'].count('A') + data['D'].count('B') < avg + 30)


# ------------------------------------------------------------------------------
#
def disabled_test_zmq_queue_cb():
    '''
    same test, but use subscriber callbacks for message delivery
    '''

    data = {'put': dict(),
            'get': dict()}
    c_a  = 2
    c_b  = 4
    cfg  = ru.Config(cfg={'uid'      : 'test_queue',
                          'channel'  : 'test',
                          'kind'     : 'queue',
                          'log_level': 'error',
                          'path'     : '/tmp/',
                          'sid'      : 'test_sid',
                          'bulk_size': 0,
                          'stall_hwm': 1,
                         })

    def get_msg_a(msg):
        uid, _ = msg.split('.')
        if uid not in data['get']:
            data['get'][uid] = list()
        data['get'][uid].append(uid)

    def get_msg_b(msg):
        uid, _ = msg.split('.')
        if uid not in data['get']:
            data['get'][uid] = list()
        data['get'][uid].append(uid)

    b = ru.zmq.Queue(cfg)
    b.start()

    assert(b.addr_in  != b.addr_out)
    assert(b.addr_in  == b.addr_put)
    assert(b.addr_out == b.addr_get)

    time.sleep(2.0)

    g_1 = ru.zmq.Getter(channel=cfg['channel'], url=str(b.addr_get), cb=get_msg_a)
    g_2 = ru.zmq.Getter(channel=cfg['channel'], url=str(b.addr_get), cb=get_msg_b)

    time.sleep(2.0)

    A = ru.zmq.Putter(channel=cfg['channel'], url=str(b.addr_put))
    B = ru.zmq.Putter(channel=cfg['channel'], url=str(b.addr_put))

    def work_put(putter, uid, n, delay):

        data['put'][uid] = list()
        idx   = 0
        while idx < n:
            time.sleep(delay)
            msg = '%s.%d' % (uid,idx)
            putter.put(msg)
            idx += 1
            data['put'][uid].append(uid)

    t_a = mt.Thread(target=work_put, args=[A, 'A', c_a, 0.010])
    t_b = mt.Thread(target=work_put, args=[B, 'B', c_b, 0.005])

    t_a.daemon = True
    t_b.daemon = True

    t_a.start()
    t_b.start()

    time.sleep(2.0)
    b.stop()

    g_1.stop()
    g_2.stop()

  # import pprint
  # pprint.pprint(data)
  #
  # uids = list(data.keys())
  # for x in uids:
  #     for y in uids:
  #         print('%s: %s: %d' % (x, y, data[x].count(y)))
  #
  # print(len(data['A']))
  # print(len(data['B']))
  # print(len(data['C']))
  # print(len(data['D']))

    assert(data['put']['A'].count('A') == c_a)
    assert(data['put']['B'].count('B') == c_b)
    assert(len(data['put']['A'])       == c_a)
    assert(len(data['put']['B'])       == c_b)

  # print(data['get']['A'].count('A'))
  # print(data['get']['B'].count('B'))
  # print(c_a)
  # print(c_b)

    assert(data['get']['A'].count('A') + data['get']['B'].count('B') == c_a + c_b)

    avg = (c_a + c_b) / 2
    assert(avg - 5 < data['get']['A'].count('A') + data['get']['B'].count('B') < avg + 5)
    assert(avg - 5 < data['get']['A'].count('A') + data['get']['B'].count('B') < avg + 5)


# ------------------------------------------------------------------------------
#
def test_zmq_queue_cb():
    '''
    same test, but use subscriber callbacks for message delivery, and only use
    one subscriber
    '''

    data = {'put': dict(),
            'get': dict()}
    c_a  = 2
    c_b  = 4
    cfg  = ru.Config(cfg={'uid'      : 'test_queue',
                          'channel'  : 'test',
                          'kind'     : 'queue',
                          'log_level': 'error',
                          'path'     : '/tmp/',
                          'sid'      : 'test_sid',
                          'bulk_size': 0,
                          'stall_hwm': 1,
                         })

    def get_msg_a(msgs):
        for msg in msgs:
            uid, _ = msg.split('.')
            if uid not in data['get']:
                data['get'][uid] = list()
            data['get'][uid].append(uid)

    b = ru.zmq.Queue(cfg)
    b.start()

    assert(b.addr_in  != b.addr_out)
    assert(b.addr_in  == b.addr_put)
    assert(b.addr_out == b.addr_get)

    ru.zmq.Getter(channel=cfg['channel'], url=str(b.addr_get), cb=get_msg_a)

    time.sleep(1.0)

    A = ru.zmq.Putter(channel=cfg['channel'], url=str(b.addr_put))
    B = ru.zmq.Putter(channel=cfg['channel'], url=str(b.addr_put))

    def work_put(putter, uid, n, delay):

        data['put'][uid] = list()
        idx   = 0
        while idx < n:
            time.sleep(delay)
            msg = '%s.%d' % (uid,idx)
            putter.put(msg)
            idx += 1
            data['put'][uid].append(uid)

    t_a = mt.Thread(target=work_put, args=[A, 'A', c_a, 0.02])
    t_b = mt.Thread(target=work_put, args=[B, 'B', c_b, 0.01])

    t_a.daemon = True
    t_b.daemon = True

    t_a.start()
    t_b.start()

    time.sleep(1.0)
    b.stop()

  # import pprint
  # pprint.pprint(data)

    assert(data['put']['A'].count('A') == c_a)
    assert(data['put']['B'].count('B') == c_b)
    assert(len(data['put']['A'])       == c_a)
    assert(len(data['put']['B'])       == c_b)

  # print(data['get']['A'].count('A'))
  # print(data['get']['B'].count('B'))
  # print(c_a)
  # print(c_b)

    assert(data['get']['A'].count('A') + data['get']['B'].count('B') == c_a + c_b)


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == '__main__':

    test_zmq_queue()
    test_zmq_queue_cb()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__author__    = "Radical.Utils Development Team (Andre Merzky)"
__copyright__ = "Copyright 2013, RADICAL@Rutgers"
__license__   = "MIT"

import os

import radical.utils as ru

# test environment specific
BLACKLIST_LOCAL_SYS = ['XPC_SERVICE_NAME']


# ------------------------------------------------------------------------------
#
def test_env_prep():

    try   : del(os.environ['TEST'])
    except: pass

    # clean `os.environ`
    for env_name in BLACKLIST_LOCAL_SYS:
        if env_name in os.environ:
            del os.environ[env_name]

    env = dict(os.environ)
    ret = ru.env_prep(env)
    only_env, only_ret, changed = ru.env_diff(env, ret)
    assert(not only_ret), only_ret
    assert(not changed), changed

    env = dict(os.environ)

    os.environ['BASH_FUNC_foo%%']  = '() {echo foo}'
    os.environ['OS_ONLY']  = 'os_only'
    os.environ['SHARED']   = 'os_shared'
    env       ['SHARED']   = 'env_shared'
    env       ['ENV_ONLY'] = 'env_only'

    ret = ru.env_prep(environment=env, unset=['OS_ONLY'], script_path='/tmp/test.env')
    assert('OS_ONLY' not in ret), ret['OS_ONLY']
    assert(ret['ENV_ONLY'] == 'env_only')
    assert(ret['SHARED'] == 'env_shared')

    only_env, only_ret, changed = ru.env_diff(env, ret)
    assert(not only_ret), [only_env, only_ret, changed]
    assert(not changed),  changed

    out, err, ret = ru.sh_callout('export OS_ONLY=x; . /tmp/test.env; echo $OS_ONLY',
                                 shell=True)
    import pprint
    print('=== ret')
    pprint.pprint(ret)
    print('=== out')
    pprint.pprint(out)
    print('=== err')
    pprint.pprint(err)
    assert(not out.strip()), out
    assert(not err.strip()), err
    assert(not ret), ret

    out, _, ret = ru.sh_callout('unset ENV_ONLY; . /tmp/test.env; echo $ENV_ONLY',
                                  shell=True)
    out = out.strip()
    assert(out == 'env_only'), out
    assert(not ret), ret


# ------------------------------------------------------------------------------
#
def test_env_read():

    fname = '/tmp/env.%d' % os.getpid()
    os.system('/bin/bash -c "env > %s"' % fname)
    try:
        env = ru.env_read(fname)

        for k,v in env.items():
            assert(os.environ[k] == v), [k, os.environ[k], v]

        for k,v in os.environ.items():
            if k not in ru.env.BLACKLIST:
                if k.startswith('BASH_FUNC_'):
                    # bash funcs are not exported to other shells
                    continue
                assert(env[k] == v)

    finally:
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
def test_env_write():

    fname     = '/tmp/env.%d' % os.getpid()
    env       = {'TEST_ENV': 'test_env'}
    unset_env = ['UNSET_ENV', '1_NON_VALID']
    try:
        ru.env_write(fname, env, unset_env)

        with open(fname) as fd:
            file_content = ''.join(fd.readlines())

        assert ("export TEST_ENV='test_env'" in file_content)
        assert ('UNSET_ENV'                  in file_content)
        assert ('1_NON_VALID'            not in file_content)
        assert ('# pre_exec'             not in file_content)

    finally:
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
def test_env_proc():

    key = 'TEST_SHARED'
    val = 'env_shared'

    env = {key: val}
    env_proc = ru.EnvProcess(env=env)

    with env_proc:
        env_proc.put(ru.sh_callout('echo -n $%s' % key, shell=True))
    out = str(env_proc.get())

    assert(isinstance(out, str))
    assert(key not in os.environ)
    assert(env[key] in out)


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test_env_prep()
    test_env_read()
    test_env_write()
    test_env_proc()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = "Radical.Utils Development Team"
__copyright__ = "Copyright 2016, RADICAL@Rutgers"
__license__   = "MIT"


'''
Unit tests for `ru.Dameon()` and `ru.daemonize()`
'''

import os
import time
import pytest

import radical.utils as ru


# ------------------------------------------------------------------------------
#
def test_daemonize():
    '''
    start a damon which sleeps for 2 seconds
    '''

    def main():
        time.sleep(2.0)

    pid = ru.daemonize(main=main)

    assert(pid)
    assert(os.kill(int(pid), 0) is None)  # process should exist

    time.sleep(3)
    with pytest.raises(OSError):
        os.kill(pid, 0)  # process should be gone

    # cleanup
    try   : os.kill(pid, 9)
    except: pass


# ------------------------------------------------------------------------------
#
def test_daemon_class():
    '''
    start a damon which sleeps for 2 seconds
    '''

    # --------------------------------------------------------------------------
    class P(ru.Daemon):

        def __init__(self):
            ru.Daemon.__init__(self, stdin=None, stdout=None, stderr=None)

        def run(self):
            time.sleep(2.0)
    # --------------------------------------------------------------------------

    p = P()
    p.start()
    assert(p.pid)
    assert(os.kill(p.pid, 0) is None)  # process should exist

    time.sleep(3)
    with pytest.raises(OSError):
        os.kill(p.pid, 0)  # process should be gone

    # cleanup
    try   : p.stop()
    except: pass


# ------------------------------------------------------------------------------
# run tests if called directly
if __name__ == "__main__":

    test_daemonize()
    test_daemon_class()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import os
import time
import pytest

import radical.utils as ru


yaml   = pytest.importorskip('yaml')
flux   = pytest.importorskip('flux')
events = dict()
spec   = {
             "tasks": [{
                 "slot": "task",
                 "count": {
                     "per_slot": 1
                 },
                 "command": [
                     "/bin/date"
                 ]
             }],
             "attributes": {
                 "system": {
                     "duration": 10000
                 }
             },
             "version": 1,
             "resources": [{
                 "count": 1,
                 "type" : "slot",
                 "label": "task",
                 "with": [{
                     "count": 1,
                     "type": "core"
                 }]
             }]
         }


# ------------------------------------------------------------------------------
#
def test_flux_startup():

    global events

    njobs    = 10
    events   = dict()

    def cb1(job_id, state, ts, context):

      # print([job_id, state, ts, context])
        if job_id not in events:
            events[job_id] = [ts, state]
        else:
            events[job_id].append([ts, state])


    fh = ru.FluxHelper()
    fh.start_flux()

    assert(fh.uri)
    assert('FLUX_URI' in fh.env)

    specs = [spec] * njobs
    ids   = fh.submit_jobs(specs, cb=cb1)
    assert(len(ids) == njobs), len(ids)

    time.sleep(5)

    assert(len(events) == njobs), len(events)
    for jid in events:
        # we expect at least 4 events per job:
        # 'submit', 'start', 'finish', 'clean',
        assert(len(events[jid]) >= 4), [jid, events[jid]]

    fh.reset()
    assert(fh.uri is None)


# ------------------------------------------------------------------------------
#
def test_flux_pickup():

    global events

    njobs    = 10
    events   = dict()
    outer_fh = None

    if 'FLUX_URI' not in os.environ:
        outer_fh = ru.FluxHelper()
        outer_fh.start_flux()

        for k,v in outer_fh.env.items():
            os.environ[k] = v

    def cb1(job_id, state, ts, context):

      # print([job_id, state, ts, context])
        if job_id not in events:
            events[job_id] = [ts, state]
        else:
            events[job_id].append([ts, state])

    fh = ru.FluxHelper()
    fh.start_flux()

    assert(fh.uri)
    assert('FLUX_URI' in fh.env)

    specs = [spec] * njobs
    ids   = fh.submit_jobs(specs, cb=cb1)
    assert(len(ids) == njobs), len(ids)

    time.sleep(5)

    assert(len(events) == njobs), len(events)
    for jid in events:
        # we expect at least 4 events per job:
        # 'submit', 'start', 'finish', 'clean',
        assert(len(events[jid]) >= 4), [jid, events[jid]]

    fh.reset()
    assert(fh.uri is None)

    if outer_fh:
        outer_fh.reset()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test_flux_startup()
    test_flux_pickup()


# ------------------------------------------------------------------------------




#!/usr/bin/env python

__author__    = "Andre Merzky"
__copyright__ = "Copyright 2012-2013, The SAGA Project"
__license__   = "MIT"


import os
import copy
import time

import radical.utils as ru

# create a virgin env
old_env = copy.deepcopy(os.environ)
for _key in list(os.environ.keys()):
    if _key.startswith('RADICAL_'):
        del(os.environ[_key])


# ------------------------------------------------------------------------------
#
def _cmd(cmd):

    _, _, ret = ru.sh_callout(cmd)

    if ret == 0: return True
    else       : return False


# ------------------------------------------------------------------------------
#
def test_profiler():
    '''
    create and check profile timestamps
    '''

    pname = 'ru.%d'        % os.getpid()
    fname = '/tmp/%s.prof' % pname
    now   = time.time()

    try:
        os.environ['RADICAL_PROFILE'] = 'True'
        prof = ru.Profiler(name=pname, ns='radical.utils', path='/tmp/')

        prof.prof('foo')
        prof.prof('bar', uid='baz')
        prof.prof('buz', ts=now)
        prof.flush()

        assert(os.path.isfile(fname))

        def _grep(pat):
            return _cmd('grep -e "%s" %s' % (pat, fname))

        assert(_grep('^[0-9\\.]*,foo,%s,MainThread,,,$'    %       pname ))
        assert(_grep('^[0-9\\.]*,bar,%s,MainThread,baz,,$' %       pname ))
        assert(_grep('^%.7f,buz,%s,MainThread,,,$'         % (now, pname)))

    finally:
        try   : del(os.environ['RADICAL_PROFILE'])
        except: pass
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
def test_enable():

    pname = 'ru.%d'        % os.getpid()
    fname = '/tmp/%s.prof' % pname

    try:
        os.environ['RADICAL_PROFILE'] = 'True'
        prof = ru.Profiler(name=pname, ns='radical.utils', path='/tmp/')

        prof.prof('foo')
        prof.disable()
        prof.prof('bar')
        prof.enable()
        prof.prof('buz')
        prof.flush()

        assert(os.path.isfile(fname))

        def _grep(pat):
            return _cmd('grep -e "%s" %s' % (pat, fname))

        assert     _grep('foo')
        assert not _grep('bar')
        assert     _grep('buz')

    finally:
        try   : del(os.environ['RADICAL_PROFILE'])
        except: pass
        try   : os.unlink(fname)
        except: pass


# ------------------------------------------------------------------------------
#
def test_env():
    '''
    Print out some messages with different log levels
    '''

    # --------------------------------------------------------------------------
    #
    def _assert_profiler(key, val, res):

        try:
            os.environ[key] = val

            pname = 'ru.%d'        % os.getpid()
            fname = '/tmp/%s.prof' % pname
            prof  = ru.Profiler(name=pname, ns='radical.utils.test',
                                path='/tmp/')
            prof.prof('foo')
            prof.flush()

            assert(res == os.path.isfile(fname))
            assert(res == _cmd('grep -e "^[0-9\\.]*,foo,%s," %s'
                              % (pname, fname)))

        finally:
            try   : del(os.environ[key])
            except: pass
            try   : os.unlink(fname)
            except: pass


    # --------------------------------------------------------------------------
    #
    for key in ['RADICAL_PROFILE',
                'RADICAL_UTILS_PROFILE',
                'RADICAL_UTILS_TEST_PROFILE']:

        for k in list(os.environ.keys()):
            if k.startswith('RADICAL'):
                del(os.environ[k])

        _assert_profiler('', '', True)

        for val, res in [
                         ['false', False],
                         ['',      True ],
                         ['1',     True ],
                         ['true',  True ],
                         ['True',  True ],
                         ['TRUE',  True ],
                         ['false', False],
                         ['False', False],
                         ['FALSE', False],
                         ['0',     False]]:

            for k in list(os.environ.keys()):
                if k.startswith('RADICAL'):
                    del(os.environ[k])

            os.environ[key] = val
            _assert_profiler(key, val, res)


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    test_profiler()
    test_env()
    test_enable()


# ------------------------------------------------------------------------------

